import argparse
import os
from pathlib import Path
import shlex

from .__main__ import Runner
from .common import Printer

SCRIPT_DIRECTORY = os.path.dirname(os.path.realpath(os.path.join(__file__)))

def mocodo(arg_string="", source=""):
    """
    Call `mocodo` as a function.
    When no source path is provided, the given source is used, and written to the output directory
    as "sandbox.mcd".
    When no source is provided, either as a file path or a string, the pristine sandbox is used as
    a fallback.

    Args:
        arg_string (str, optional): The same arguments as the command line. Defaults to "".
        source (str, optional): The source code of the MCD file. Defaults to "".
    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("--input", "-i")
    parser.add_argument("--output_dir")
    (args, remaining_args) = parser.parse_known_args(shlex.split(arg_string))
    input_dir = Path.cwd()

    # Create or retrieve the path to the source file.
    if not args.input:
        # No path is provided for the source of the MCD.
        if source:
            # The source is provided as a string. Write it to a file named "sandbox.mcd".
            input_path = input_dir / "sandbox.mcd"
            input_path.write_text(source, encoding="utf8")
        else:
            # No source is provided. Fall back to the pristine sandbox.
            input_path = Path(SCRIPT_DIRECTORY, "resources", "pristine_sandbox.mcd")
    else:
        # Retrieve the path to the source file.
        input_path = Path(args.input)
        if not input_path.suffix:
            input_path = input_path.with_suffix(".mcd")

    if not args.output_dir:
        # No output directory is provided. Fall back to the input directory.
        output_dir = input_dir
    else:
        # Retrieve the path to the output directory, and create it if it doesn't exist.
        output_dir = Path(args.output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    remaining_args.extend([
        "--input", str(input_path),
        "--output_dir", str(output_dir)
    ])

    printer = Printer(accumulate=True) # silence the success messages
    run = Runner(remaining_args, printer)
    run()



# http://mwh.geek.nz/2009/04/26/python-damerau-levenshtein-distance/


def distance(seq1, seq2):
    """Calculate the Damerau-Levenshtein distance between sequences.

    This distance is the number of additions, deletions, substitutions,
    and transpositions needed to transform the first sequence into the
    second. Although generally used with strings, any sequences of
    comparable objects will work.

    Transpositions are exchanges of *consecutive* characters; all other
    operations are self-explanatory.

    This implementation is O(N*M) time and O(M) space, for N and M the
    lengths of the two sequences.

    >>> damerau_levenshtein('ba', 'abc')
    2
    >>> damerau_levenshtein('fee', 'deed')
    2
    It works with arbitrary sequences too:
    >>> damerau_levenshtein('abcd', ['b', 'a', 'c', 'd', 'e'])
    2
    """
    # codesnippet:D0DE4716-B6E6-4161-9219-2903BF8F547F
    # Conceptually, this is based on a len(seq1) + 1 * len(seq2) + 1 matrix.
    # However, only the current and two previous rows are needed at once,
    # so we only store those.
    one_ago = None
    this_row = list(range(1, len(seq2) + 1)) + [0]
    for x in range(len(seq1)):
        # Python lists wrap around for negative indices, so put the
        # leftmost column at the *end* of the list. This matches with
        # the zero-indexed strings and saves extra calculation.
        two_ago, one_ago, this_row = one_ago, this_row, [0] * len(seq2) + [x + 1]
        for y in range(len(seq2)):
            del_cost = one_ago[y] + 1
            add_cost = this_row[y - 1] + 1
            sub_cost = one_ago[y - 1] + (seq1[x] != seq2[y])
            this_row[y] = min(del_cost, add_cost, sub_cost)
            # This block deals with transpositions
            if (x > 0 and y > 0 and seq1[x] == seq2[y - 1]
                and seq1[x-1] == seq2[y] and seq1[x] != seq2[y]):
                this_row[y] = min(this_row[y], two_ago[y - 2] + 1)
    return this_row[len(seq2) - 1]



import json
import numbers
import sys
from pathlib import Path
import contextlib
import importlib

from .mocodo_error import MocodoError


def contract_path(path):
    """
    Contract a full path to one with ~, if it's under the user's home directory.
    """
    home = str(Path.home())
    absolute_path = str(Path(path).resolve())
    if absolute_path.startswith(home):
        return absolute_path.replace(home, '~', 1)
    else:
        return absolute_path


class Printer:

    def __init__(self, accumulate=False):
        self.accumulator = []
        if accumulate:
            self.write = self.accumulator.append
        else:
            self.write = self.safe_print_for_PHP

    def safe_print_for_PHP(self, s):
        """ It seems that when called from PHP, Python is unable to guess correctly
            the encoding of the standard output. """
        try:
            print(s, file=sys.stdout)
        except UnicodeEncodeError:
            print(s.encode("utf8"), file=sys.stdout)


class Common:

    def __init__(self, params):
        self.params = params

    def output_success_message(self, path):
        return _('Output file "{filename}" successfully generated.').format(filename=contract_path(path))

    def update_source(self, source):
        path = Path(f"{self.params['output_name']}.mcd")
        path.write_text(source, encoding="utf8")
        return _('Source file "{filename}" successfully updated.').format(filename=contract_path(path))

    def load_input_file(self):
        # Try to read the file locally.
        path = Path(self.params["input"])
        if path.is_file():
            for encoding in self.params["encodings"]:
                with contextlib.suppress(UnicodeError):
                    self.encoding = encoding
                    return path.read_text(encoding=encoding).replace('"', '')
            raise MocodoError(5, _('Unable to read "{filename}" with any of the following encodings: "{encodings}".').format(filename=self.params["input"], encodings= ", ".join(self.params["encodings"]))) # fmt: skip
        # The file is not found locally, try to retrieve it from the server.
        self.encoding = self.params["encodings"][0]
        requests = importlib.import_module("requests")
        with contextlib.suppress(requests.exceptions.ConnectionError):
            response = requests.get(f"{self.params['lib']}/{path.name}")
            if response.status_code == 200:
                source = response.text.replace('"', '')
                # Save the file locally.
                path.parent.mkdir(parents=True, exist_ok=True)
                path.write_text(source, encoding=self.encoding)
                return source
        # The file is not found locally nor on the server.
        raise MocodoError(2, _('The file "{input}" doesn\'t exist.').format(input=path))  # fmt: skip

    def load_style(self):
        
        def load_by_name(name):
            path = Path(self.params[name]).with_suffix(".json")
            if path.is_file():
                try:
                    return json.loads(path.read_text(encoding="utf8"))
                except:
                    raise MocodoError(3, _('Problem with "{name}" file "{path}".').format(name=name, path=path)) # fmt: skip
            path = Path(self.params["script_directory"], "resources", name, path)
            try:
                return json.loads(path.read_text(encoding="utf8"))
            except:
                raise MocodoError(3, _('Problem with "{name}" file "{path}".').format(name=name, path=path)) # fmt: skip
        
        def may_apply_scaling(shapes):
            if self.params["scale"] == 1:
                return
            scale = self.params["scale"]
            for key in shapes:
                if key.endswith("font"):
                    shapes[key]["size"] = shapes[key]["size"] * scale
                elif not key.endswith("ratio") and isinstance(shapes[key], numbers.Number):
                    shapes[key] *= scale
        
        def ensure_margins_are_integer(shapes):
            # Some nasty failures are known to occur otherwise.
            for key in shapes:
                if "margin" in key:
                    shapes[key] = int(shapes[key] + 0.5)
            
        style = {}
        style.update(load_by_name("colors"))
        shapes = load_by_name("shapes")
        may_apply_scaling(shapes)
        ensure_margins_are_integer(shapes)
        style.update(shapes)
        style["transparent_color"] = None
        return style



__import__("sys").path[0:0] = ["."]

import re

from ..tools.string_tools import markdown_table
from ..parse_mcd import Transformer
from ..tools.parser_tools import parse_source, first_child
from ..mocodo_error import subsubopt_error

DEFAULT_COLUMN_NAMES = {
    "en": {
        "label": "Label of the attribute",
        "type": "Type",
        "box": "Entity or relationship",
    },
    "fr": {
        "label": "Libellé de l'attribut",
        "type": "Type",
        "box": "Entité ou association",
    },
}

class AttributeListExtractor(Transformer): # depth-first, post-order
    def __init__(self):
        self.boxes = {}
        self.typed_attribute_accumulator = []
        self.assoc_clause = self._box_clause
        self.entity_clause = self._box_clause
    
    def _box_clause(self, children):
        # Associate the box name with the accumulated list of attributes
        box_def_prefix = first_child(children[0], "box_def_prefix")
        if box_def_prefix != "-":
            if not box_def_prefix:
                box_name = first_child(children[0], "box_name")
            else:
                box_name = first_child(children[1], "box_name")
            self.boxes[box_name] = self.typed_attribute_accumulator[:]
        self.typed_attribute_accumulator = []
    
    def attr(self, children):
        # Accumulate a couple (attribute name, data type placeholder)
        self.typed_attribute_accumulator.append((children[0].value, ""))
    
    def datatype(self, children):
        # Replace the last data type placeholder with the actual data type
        (name, _) = self.typed_attribute_accumulator.pop()
        self.typed_attribute_accumulator.append((name, children[0].value))
    
    def finalize(self, common, subargs):
        language = common.params["language"]
        actual_colons = {}
        for kind in list(subargs):
            m = re.match(r"^([*_`]*)(.+)\1$", kind) 
            if m[2] in ("label", "type", "box"):
                actual_colons[m[2]] = m[1]
            subargs[m[2]] = subargs.pop(kind)
        if not actual_colons:
            actual_colons["box"] = subargs["box"] = ""
            actual_colons["label"] = subargs["label"] = ""
            actual_colons["type"] = subargs["type"] = ""
        self.header = []
        self.projectors = []
        for (kind, col_name) in subargs.items():
            if kind in ("tsv", "md", "markdown"):
                continue
            if kind not in actual_colons:
                raise subsubopt_error(kind)
            self.header.append(col_name or DEFAULT_COLUMN_NAMES[language][kind])
            self.projectors.append(kind)
        self.rows = []
        for (box, attributes) in self.boxes.items():
            for (attr, datatype) in attributes:
                d = {"box": box, "label": attr, "type": datatype}
                row = [d[p] for p in self.projectors]
                self.rows.append(row)
        self.md_tags = [actual_colons[p] for p in self.projectors]
        self.rows.sort()
        return len(actual_colons)
    
    def get_tsv(self):
        result = []
        if len(self.header) > 1:
            result.append("\t".join(self.header))
        result.extend(map("\t".join, self.rows))
        return "\n".join(result)
    
    def get_markdown(self):
        if len(self.header) > 1: # output a table only if there are at least two columns
            result = []
            previous_leftmost = None
            for row in [self.header] + self.rows:
                if row[0] == previous_leftmost:
                    row[0] = '"'
                else:
                    previous_leftmost = row[0]
                result.append(f"{tag}{cell}{tag}" for (tag, cell) in zip(self.md_tags, row))
            return markdown_table(result)
        else: # otherwise, output a list
            return "\n".join([f"- {self.md_tags[0]}{row[0]}{self.md_tags[0]}" for row in self.rows])

def run(source, subargs=None, common=None):
    tree = parse_source(source)
    extractor = AttributeListExtractor()
    extractor.transform(tree)
    column_count = extractor.finalize(common, subargs)
    if "tsv" in subargs:
        return {
            "stem_suffix": f"_data_dict_{column_count}",
            "text": extractor.get_tsv(),
            "extension": "tsv",
            "to_defer": False,
            "highlight": "tsv",
        }
    else: # Markdown is the default
        return {
            "stem_suffix": f"_data_dict_{column_count}",
            "text": extractor.get_markdown(),
            "extension": "md",
            "to_defer": False,
            "highlight": "markdown",
        }



from collections import Counter

__import__("sys").path[0:0] = ["."]

from ..parse_mcd import Token, Visitor
from ..tools.parser_tools import parse_source, reconstruct_source, first_child

class Splitter(Visitor):

    def line(self, tree):
        # It is not possible to use a method `assoc_clause` since the amount of indentation needs
        # to be known for the added clauses.

        # Guard: store the indentation and abort if the clause is not an association.
        indent = next(tree.find_data("indent"), None)
        indent = indent.children[0].value if indent else ""
        tree = next(tree.find_data("assoc_clause"), None)
        if tree is None:
            return

        # Guard: abort if the association has less than 3 legs.
        legs = [node for node in tree.find_data("assoc_leg")]
        if len(legs) < 3:
            return

        # Reorder the legs with "11" cards first, then "01", then the other ones.
        d = {"11": 0, "01": 1}
        legs.sort(key=lambda leg: d.get(first_child(leg, "card").value, 2))

        # Guard: ensure that the first max cardinality is 1.
        if first_child(legs[0], "card")[1] != "1":
            return
        
        # Guard: don't split a clustered association
        card_prefixes = [first_child(leg, "card_prefix") == "/" for leg in legs]
        if any(card_prefixes):
            return
        
        # Accumulate the constituting elements (cf. grammar.lark) of the original legs.
        new_items = []
        for leg in legs:
            acc = []
            acc.append(first_child(leg, "card_hidden"))
            acc.append(first_child(leg, "card_prefix"))
            acc.append(first_child(leg, "card"))
            acc.append(first_child(leg, "leg_arrow"))
            acc.append(first_child(leg, "leg_note"))
            acc[-1] = f" [{acc[-1]}] " if acc[-1] else " "
            acc.append(first_child(leg, "entity_name_ref").children[0])
            new_items.append("".join(acc))
        
        # Concatenate any attributes of the association.
        attrs = []
        for child in tree.find_data("typed_attr"):
            acc = []
            acc.append(first_child(child, "attr"))
            datatype  = first_child(child, "datatype")
            if datatype:
                acc.append(f" [{datatype}]")
            attrs.append("".join(acc))
        
        # Construct the lines of the exploded association. The *1 leg is the first one.
        result = []
        box_def_prefix = first_child(tree, "box_def_prefix")
        assoc_name_def = first_child(tree, "assoc_name_def").children[0]
        for (i, new_item) in enumerate(new_items[1:]):
            result.append(f"{indent}{box_def_prefix}{assoc_name_def}{i}, {new_items[0]}, {new_item}")

        # Add the attributes to the first clause only. Duplicating them would be an error.
        result[0] += f": {', '.join(attrs)}" if attrs else ""

        # The first clause starts at the same position as the association name: unindent.
        result[0] = result[0].lstrip()

        # Add a blank line after the exploded association.
        result.append("")

        # Replace the original tree by a pseudo-tree containing only the desired string.
        token = first_child(tree, "box_name")
        tree.children = [Token("MOCK", "\n".join(result), line=token.line, column=token.column)]

def run(source, **kargs):
    tree = parse_source(source)
    visitor = Splitter()
    visitor.visit(tree)
    return reconstruct_source(tree)



import contextlib
from pathlib import Path
import random
import re

__import__("sys").path[0:0] = ["."]

from ..parse_mcd import Transformer
from ..tools.parser_tools import transform_source
from ..tools.string_tools import ascii, camel, snake, pascal, TRUNCATE_DEFAULT_SIZE
from ..mocodo_error import MocodoError
from .cards import fix_card, infer_dfs, infer_roles
from .types import read_default_datatypes, create_type_placeholders, guess_types
from .obfuscate import obfuscator_factory
from .arrows import create_df_arrows
from .constraints import create_cifs

ELEMENT_TO_TOKENS = {
    "arrows": ["leg_arrow"],
    "attrs": ["attr"],
    "boxes": ["box_name"],
    "card_prefixes": ["card_prefix"],
    "cards": ["card"],
    "roles": ["leg_note"],
    "leg_notes": ["leg_note"],
    "constraint_notes": ["constraint_note"],
    "labels": ["box_name", "attr"],
    "texts": ["box_name", "attr", "leg_note", "constraint_note"],
    "notes": ["leg_note", "constraint_note"],
    "types": ["datatype"],
}

GENERAL_OPERATIONS = { # operations that can be applied to any token
    "ascii": ascii,
    "camel": camel,
    "capitalize": lambda x: x.capitalize(),
    "casefold": lambda x: x.casefold(),
    "echo": lambda x: x,
    "lower": lambda x: x.lower(),
    "pascal": pascal,
    "snake": snake,
    "swapcase": lambda x: x.swapcase(),
    "title": lambda x: x.title(),
    "upper": lambda x: x.upper(),
}


class Mapper(Transformer):

    def __init__(self, op_name, pre_token, subsubarg, params):
        tokens = ELEMENT_TO_TOKENS[pre_token]
        op = GENERAL_OPERATIONS.get(op_name)
        if op is None: # op_tk operations with limited applicability and/or using a subsubarg
            if op_name == "randomize" and pre_token == "types":
                resource_dir = Path(params["script_directory"], "resources")
                pool = list(dict(read_default_datatypes(resource_dir)).values())
                op = lambda _: random.choice(pool)
            elif op_name == "delete" and pre_token in ("attrs", "notes", "roles", "constraint_notes", "arrows", "types", "card_prefixes"):
                op = lambda _: ""
            elif op_name == "delete" and pre_token == "cards":
                op = lambda _: "XX"
            elif op_name == "fix" and pre_token == "cards":
                op = fix_card
            elif op_name == "randomize" and pre_token in ("labels", "texts", "boxes", "attrs", "notes", "roles", "constraint_notes"):
                op = obfuscator_factory(subsubarg, params)
            elif op_name == "truncate":
                truncate_size = TRUNCATE_DEFAULT_SIZE
                if subsubarg and subsubarg.isdigit():
                    truncate_size = int(subsubarg) or truncate_size
                op = lambda x: x[:truncate_size]
            elif op_name == "slice":
                slice_start = slice_stop = None
                if subsubarg:
                    with contextlib.suppress(ValueError):
                        slice_start = int(subsubarg.partition(":")[0])
                    with contextlib.suppress(ValueError):
                        slice_stop = int(subsubarg.partition(":")[2])
                op = lambda x: x[slice_start:slice_stop]
            elif op_name == "replace":
                (substring, __, repl) = subsubarg.partition("/")
                op = lambda x: x.replace(substring, repl)
            elif op_name == "suffix":
                op = lambda x: f"{x}{subsubarg}"
            elif op_name == "prefix":
                op = lambda x: f"{subsubarg}{x}"
            else:
                raise MocodoError(24, _('Operation "{op_name}" cannot be applied to "{pre_token}".').format(op_name=op_name, pre_token=pre_token))
        update_first_child = lambda tree: tree[0].update(value=op(tree[0].value))
        for token in tokens:
            setattr(self, token, update_first_child)


def run(source, op_name, subargs, params, **kargs):
    if op_name == "randomize" and not subargs:
        subargs = {"labels": ""} # used for obfuscation
    elif op_name == "delete" and not subargs:
        subargs = dict.fromkeys("types notes attrs cards".split(), "")
    for (pre_token, subsubarg) in subargs.items():
        # filter special non-op_tk operations
        if op_name == "create" and pre_token == "types":
            source = create_type_placeholders(source, subsubarg) if subsubarg is not None else guess_types(source, params)
        elif op_name == "create" and pre_token == "df_arrows":
            source = create_df_arrows(source, subsubarg)
        elif op_name == "create" and re.match("(?i)dfs?$", pre_token):
            source = infer_dfs(source, params["df"])
        elif op_name == "create" and pre_token == "roles":
            source = infer_roles(source)
        elif op_name == "create" and re.match("(?i)cifs?$", pre_token):
            source = create_cifs(source, subsubarg)
        else: # apply a normal op_tk operation
            source = transform_source(source, Mapper(op_name, pre_token, subsubarg, params))
        if op_name == "delete":
            # After a delete operation, remove any [] that may have been left behind
            # (it's too tedious to remove them from the AST, and it's not worth it).
            source = re.sub(r" *\[\]", " ", source)
    return source



import re

__import__("sys").path[0:0] = ["."]

from ._crow import Crow
from ..tools.parser_tools import parse_source
from ..rewrite import (
    op_tk,
    _drain as drain,
    _explode as explode,
    _split as split,
)
from ..tools.string_tools import rstrip_digit_or_underline
from ..version_number import version

SUFFIX = "_crows_foot_erd.mmd"

LEFT_CARD = {
    "01": "|o",
    "0N": "}o",
    "11": "||",
    "1N": "}|",
}
RIGHT_CARD = {
    "01": "o|",
    "0N": "o{",
    "11": "||",
    "1N": "|{",
}


def sanitize_type(s):
    # In Mermaid syntax, the type values must begin with an alphabetic character
    # and may contain digits, hyphens, underscores, parentheses and square brackets.
    # This seems a too specific operation to be offered by op_tk.
    s = s.replace(
        ",", "-"
    )  # as long as https://github.com/mermaid-js/mermaid/issues/1546 is not fixed
    s = re.sub(r"[^-_0-9A-Za-z()[\]]", "_", s)
    s = re.sub(r"__+", "_", s)
    s = s.strip("_")
    return s


class CrowMmd(Crow):
    def get_text(self):
        result = []
        result.append("erDiagram")
        result.append(f"  %% Generated by Mocodo {version}")
        for (name, has_id, attrs) in self.tables.values():
            result.append(f"  {name} {{")
            for (datatype, attr, is_id) in attrs:
                datatype = sanitize_type(datatype) if datatype else "TYPE"
                pk = " PK" if is_id else ""
                result.append(f"    {datatype} {attr}{pk}")
            result.append(f"  }}")
        for (ent_1, card_1, kind, card_2, ent_2, assoc_name) in self.links:
            ent_1 = rstrip_digit_or_underline(ent_1)
            ent_2 = rstrip_digit_or_underline(ent_2)
            card_1 = LEFT_CARD.get(card_1, "}|")
            card_2 = RIGHT_CARD.get(card_2, "|{")
            result.append(f"  {ent_1} {card_1}{kind}{card_2} {ent_2}: {assoc_name}")
        return "\n".join(result)


def run(source, subargs, common=None):
    source = op_tk.run(source, "ascii", {"labels": 1}, common.params)
    source = op_tk.run(source, "snake", {"labels": 1}, common.params)
    source = drain.run(source)
    source = split.run(source)
    source = explode.run(source, {"arity": "2.5", "weak": True}, common.params)
    tree = parse_source(source)
    extractor = CrowMmd()
    extractor.visit(tree)
    result = extractor.get_text()
    return result



from pathlib import Path
import re
import requests

url = "https://modern-sql.com/reserved-words-empirical-list.html"
text = requests.get(url).text

words = {}
for (y, word) in re.findall(r'y="(\d+)"><tspan class="code.*?">(.+?)</tspan>', text):
    words[str(int(y) - 14)] = word

words_by_dialect = {}
dialect = None
for chunk in text.split('<g class="hoverable">'):
    if m := re.search(r'rotate\(-45\)">(.+?)</text>', chunk):
        dialect = m[1]
        words_by_dialect[dialect] = []
        for (href, y) in re.findall(r'<use href="#(.+?)" transform="translate\(\d+,(\d+)\)"/>', chunk):
            if href != "none":
                words_by_dialect[dialect].append(words[y])

dialects = {
    "Apache Derby": "",
    "BigQuery": "",
    "Db2 (LUW)": "",
    "H2": "",
    "MariaDB": "",
    "MySQL": "mysql.yaml",
    "Oracle DB": "oracle.yaml",
    "PostgreSQL": "postgresql.yaml",
    "SQL Server": "mssql.yaml",
    "SQLite": "sqlite.yaml",
}

for (dialect, filename) in dialects.items():
    if not filename:
        continue
    alternative = "|".join(words_by_dialect[dialect])
    path = Path("mocodo", "resources", "relation_templates", filename)
    text = path.read_text(encoding="utf8")
    text = re.sub(r"('Protect reserved keywords'\n    search: ).+", fr"\1'(?i)^({alternative})$'", text)
    path.write_text(text, encoding="utf8")
    print(f"{filename} updated.")



import re
from .attribute import *
from .leg import *
from .mocodo_error import MocodoError
from .tools.string_tools import rstrip_digit_or_underline, raw_to_bid

class Association:

    df_counter = 0

    @classmethod
    def reset_df_counter(cls):
        cls.df_counter = 0

    def __init__(self, clause, **params):
        self.source = clause["source"]
        self.raw_name = clause["name"]
        self.bid = raw_to_bid(self.raw_name)
        # A protected association results in a table, even if this association is a DF.
        self.is_protected = (clause.get("box_def_prefix") == "+")
        if clause.get("box_def_prefix") == "-":
            self.calculate_size = self.calculate_size_when_invisible
            self.description = lambda *ignored: []
            self.is_invisible = True
        else:
            self.calculate_size = self.calculate_size_when_visible
            self.description = self.description_when_visible
            self.is_invisible = False
        self.peg_count = 0
        self.attributes = []
        for attr in clause.get("attrs", []):
            if attr.get("attribute_label", "") == "":
                self.attributes.append(PhantomAttribute(attr))
            elif attr.get("id_mark", "") == "_":
                if params.get("no_assoc_ids"):
                    raise MocodoError(52, _('The association "{name}" cannot have an identifier.').format(name=self.raw_name))
                self.attributes.append(StrongAttribute(attr))
            else:
                self.attributes.append(SimpleAssociationAttribute(attr))
        df_label = params.get("df", "DF")
        if re.match(fr"^{df_label.upper()}\d*$", self.raw_name.upper()):
            self.name_view = df_label # strip all digits suffixing a DF name
            self.bid = f"{df_label}{Association.df_counter}"
            Association.df_counter += 1
            self.kind = "df"
        else:
            self.name_view = rstrip_digit_or_underline(self.raw_name)
            legs = clause["legs"]
            # A "peg" is a leg that is prefixed by a "/" character.
            for leg in legs:
                if leg.get("card_prefix") == "/":
                    self.peg_count += 1
            if self.peg_count > 0:
                if len(legs) < 3:
                    raise MocodoError(51, _('The association "{name}" should have at least 3 legs to become a cluster.').format(name=self.raw_name))
                self.kind = "cluster"
            else:
                self.kind = "association"
        self.set_view_strategies()
        entity_raw_names = [leg["entity"] for leg in clause["legs"]]
        self.legs = []
        for leg_clause in clause["legs"]:
            entity_raw_name = leg_clause["entity"]
            rank = leg_clause["rank"]
            leg = Leg(self, leg_clause, **params)
            mutiplicity = entity_raw_names.count(entity_raw_name)
            rank = entity_raw_names[:rank].count(entity_raw_name)
            leg.set_spin_strategy(0 if mutiplicity == 1 else 2 * rank / (mutiplicity - 1) - 1)
            self.legs.append(leg)
        if self.kind == "cluster":
            candidate_groups = iter([0] + list(range(self.peg_count - 1, 0, -1)))
            for leg in self.legs:
                if leg.kind == "cluster_peg":
                    group_number = str(next(candidate_groups))
                    for other_leg in self.legs:
                        if other_leg is leg:
                            continue
                        other_leg.append_candidate_group(group_number)
        elif self.kind == "df":
            for leg in self.legs:
                if leg.card.endswith(("1", "X")):
                    break
            else:
                raise MocodoError(37, _('An association named "{df_label}" must have at least one leg with a maximal cardinality of 1.').format(df_label=df_label)) # fmt: skip


    def register_boxes(self, boxes):
        self.boxes = boxes

    def register_mcd_has_cif(self, mcd_has_cif):
        self.mcd_has_cif = mcd_has_cif
        for leg in self.legs:
            leg.register_mcd_has_cif(mcd_has_cif)

    def calculate_size_when_invisible(self, *ignored):
        self.w = 0
        self.h = 0
    
    def calculate_size_when_visible(self, style, get_font_metrics):
        cartouche_font = get_font_metrics(style["association_cartouche_font"])
        self.get_cartouche_string_width = cartouche_font.get_pixel_width
        self.cartouche_height = cartouche_font.get_pixel_height()
        attribute_font = get_font_metrics(style["association_attribute_font"])
        self.attribute_height = attribute_font.get_pixel_height()
        self.calculate_size_depending_on_kind(style, get_font_metrics)
        self.w += self.w % 2
        self.h += self.h % 2
        for leg in self.legs:
            leg.calculate_size(style, get_font_metrics)

    def register_center(self, geo):
        self.cx = geo["cx"][self.bid]
        self.cy = geo["cy"][self.bid]
        self.l = self.cx - self.w // 2
        self.r = self.cx + self.w // 2
        self.t = self.cy - self.h // 2
        self.b = self.cy + self.h // 2

    def set_view_strategies(self):

        def calculate_size_when_df(style, get_font_metrics):
            self.w = self.h = max(
                style["round_rect_margin_width"] * 2 + self.get_cartouche_string_width(self.name_view),
                style["round_rect_margin_width"] * 2 + self.cartouche_height
            )

        def calculate_size_when_default(style, get_font_metrics):
            for attribute in self.attributes:
                attribute.calculate_size(style, get_font_metrics)
            cartouche_and_attribute_widths = [a.w for a in self.attributes] + [self.get_cartouche_string_width(self.name_view)]
            self.w = 2 * style["round_rect_margin_width"] + max(cartouche_and_attribute_widths)
            self.h = max(1, len(self.attributes)) * (self.attribute_height + style["line_skip_height"]) \
                - style["line_skip_height"] \
                + 2 * style["rect_margin_height"] \
                + 2 * style["round_rect_margin_height"] \
                + self.cartouche_height
            self.w += self.w % 2
            self.h += self.h % 2

        def description_when_df(style):
            return [
                (
                    "circle",
                    {
                        "stroke_depth": style["box_stroke_depth"],
                        "stroke_color": style["association_stroke_color"],
                        "color": style["association_cartouche_color"],
                        "cx": self.cx,
                        "cy": self.cy,
                        "r": self.w // 2,
                    },
                ),
                (
                    "text",
                    {
                        "text": self.name_view,
                        "text_color": style['association_cartouche_text_color'],
                        "x": self.l + style["round_rect_margin_width"],
                        "y": self.t + style["round_rect_margin_height"] + style["df_text_height_ratio"] * self.cartouche_height,
                        "family": style["association_cartouche_font"]["family"],
                        "size": style["association_cartouche_font"]["size"],
                    },
                )
            ]
        
        def optional_description_for_cluster(style):
            if self.kind != "cluster" or self.mcd_has_cif:
                return []
            if len(self.legs) > 3:
                # The calculation of a hull with more than 3 legs is not implemented yet.
                return []
            result = []
            tweak = 1
            for (i, leg) in enumerate(self.legs):
                if leg.kind == "cluster_peg":
                    other_legs = self.legs[:i] + self.legs[i+1:]
                    e1 = other_legs[0].entity # the first entity of the other legs
                    e2 = other_legs[-1].entity # the last one, either the same as e1 or not
                    points = self.cluster_hull(e1, e2, style["card_margin"] * tweak)
                    if not points: # complex hulls are not implemented yet
                        continue
                    path = points_to_rounded_path(points, style["round_corner_radius"] // 2)
                    result.extend([
                        (
                            "polygon",
                            {
                                "path": path,
                                "stroke_color": None,
                                "stroke_depth": 0,
                                "color": style["entity_color"],
                                "opacity": 0.3 / self.peg_count,
                            },
                        ),
                        (
                            "dot_polygon",
                            {
                                "path": path,
                                "stroke_color": style['entity_stroke_color'],
                                "stroke_depth": style["box_stroke_depth"],
                                "color": None,
                                "dash_gap": style["dash_width"],
                            },
                        )
                    ])
                    tweak += 1
            return result

        def description_when_default(style):
            result = []
            result.extend(optional_description_for_cluster(style))
            if self.is_protected:
                result.append(
                    (
                        "round_rect",
                        {
                            "x": self.l - style["card_margin"] * 2,
                            "y": self.t - style["card_margin"] * 2,
                            "w": self.w + style["card_margin"] * 4,
                            "h": self.h + style["card_margin"] * 4,
                            "radius": style["round_corner_radius"] // 2,
                            "stroke_color": style['entity_stroke_color'] + "55", # alpha channel
                            "stroke_depth": style["box_stroke_depth"],
                            "color": style["entity_color"] + "55", # alpha channel
                        },
                    )
                )
            x = self.l
            y = self.t
            w = self.w
            h = self.attribute_height + style["round_rect_margin_height"] + style["rect_margin_height"]
            r = style["round_corner_radius"]
            result.append(
                (
                    "upper_round_rect",
                    {
                        "x0": x + w - r,
                        "y0": y,
                        "y1": y + h,
                        "y2": y + r,
                        "w": w,
                        "r": r,
                        "color": style['association_cartouche_color'],
                        "stroke_color": style['association_cartouche_color'],
                        "stroke_depth": 0,
                    },
                )
            )
            y = self.t + self.attribute_height + style["round_rect_margin_height"] + style["rect_margin_height"]
            h = self.h - self.attribute_height - style["round_rect_margin_height"] - style["rect_margin_height"]
            result.append(
                (
                    "lower_round_rect",
                    {
                        "x0": x + w,
                        "y0": y,
                        "y1": h - r,
                        "x1": x + r,
                        "w": w,
                        "r": r,
                        "color": style['association_color'],
                        "stroke_color": style['association_color'],
                        "stroke_depth": 0,
                    },
                )
            )
            result.append(
                (
                    "round_rect",
                    {
                        "x": self.l,
                        "y": self.t,
                        "w": self.w,
                        "h": self.h,
                        "radius": r,
                        "color": style['transparent_color'],
                        "stroke_color": style['association_stroke_color'],
                        "stroke_depth": style["box_stroke_depth"],
                    },
                )
            )
            result.append(
                (
                    "line",
                    {
                        "x0": self.l,
                        "y0": self.t + self.attribute_height + style["round_rect_margin_height"] + style["rect_margin_height"],
                        "x1": self.r,
                        "y1": self.t + self.attribute_height + style["round_rect_margin_height"] + style["rect_margin_height"],
                        "stroke_color": style['association_stroke_color'],
                        "stroke_depth": style["inner_stroke_depth"],
                    },
                )
            )
            result.append(
                (
                    "text",
                    {
                        "text": self.name_view,
                        "text_color": style['association_cartouche_text_color'],
                        "x": self.cx - self.get_cartouche_string_width(self.name_view) // 2,
                        "y": self.t + style["rect_margin_height"] + style["cartouche_text_height_ratio"] * self.cartouche_height,
                        "family": style["association_cartouche_font"]["family"],
                        "size": style["association_cartouche_font"]["size"],
                    },
                ),
            )
            x = self.cx - self.w // 2 + style["round_rect_margin_width"]
            dx = 0
            dy = style["round_rect_margin_height"] + self.cartouche_height + 2 * style["rect_margin_height"] - self.h // 2
            for attribute in self.attributes:
                result.extend(attribute.description(style, x, self.cy, dx, dy))
                dy += self.attribute_height + style["line_skip_height"]
            return result

        if self.kind == "df":
            self.calculate_size_depending_on_kind = calculate_size_when_df
            self.description_depending_on_kind = description_when_df
        else:
            self.calculate_size_depending_on_kind = calculate_size_when_default
            self.description_depending_on_kind = description_when_default

    def description_when_visible(self, style, geo):
        self.saved_card_description = []
        result = []
        result.append(("comment", {"text": f"Association {self.bid}"}))
        result.append(
            (
                "begin_component",
                {
                    "page": self.page,
                    "visibility": "hidden" if self.page else "visible",
                }
            )
        )
        result.extend(self.leg_descriptions(style, geo))
        result.append(("begin_group", {}))
        result.extend(self.description_depending_on_kind(style))
        result.append(("end", {}))
        result.extend(self.saved_card_description) # need to be displayed after the clusters for more readability
        result.append(("end", {}))
        return result

    def leg_descriptions(self, style, geo):
        result = []
        for leg in self.legs:
            result.extend(leg.description(style, geo))
            self.saved_card_description.extend(leg.saved_card_description)
        return result

    def cluster_hull(self, e1, e2, card_margin):
        x_min = min(box.l for box in (self, e1, e2))
        y_min = min(box.t for box in (self, e1, e2))
        x_max = max(box.r for box in (self, e1, e2))
        y_max = max(box.b for box in (self, e1, e2))
        x = x_min - card_margin // 2
        y = y_min - card_margin // 2
        w = x_max - x_min + card_margin
        h = y_max - y_min + card_margin

        x1_min = min(box.l for box in (self, e1))
        y1_min = min(box.t for box in (self, e1))
        x1_max = max(box.r for box in (self, e1))
        y1_max = max(box.b for box in (self, e1))
        x1 = x1_min - card_margin // 2
        y1 = y1_min - card_margin // 2
        w1 = x1_max - x1_min + card_margin
        h1 = y1_max - y1_min + card_margin
        
        x2_min = min(box.l for box in (self, e2))
        y2_min = min(box.t for box in (self, e2))
        x2_max = max(box.r for box in (self, e2))
        y2_max = max(box.b for box in (self, e2))
        x2 = x2_min - card_margin // 2
        y2 = y2_min - card_margin // 2
        w2 = x2_max - x2_min + card_margin
        h2 = y2_max - y2_min + card_margin

        e1_same_row = e1.t <= self.t < self.b <= e1.b or self.t <= e1.t < e1.b <= self.b
        e2_same_row = e2.t <= self.t < self.b <= e2.b or self.t <= e2.t < e2.b <= self.b
        e1_same_col = e1.l <= self.l < self.r <= e1.r or self.l <= e1.l < e1.r <= self.r
        e2_same_col = e2.l <= self.l < self.r <= e2.r or self.l <= e2.l < e2.r <= self.r
        if e1_same_row and e2_same_row or e1_same_col and e2_same_col:
            return [x, y, x+w, y, x+w, y+h, x, y+h]
        elif e1_same_row and e2_same_col:
            if e1.cx < self.cx:
                if e2.cy < self.cy:
                    #   2
                    # 1 a
                    return [x, y1, x2, y1, x2, y, x+w, y, x+w, y+h, x, y+h]
                else:
                    # 1 a
                    #   2
                    return [x, y, x+w, y, x+w, y+h, x2, y+h, x2, y1+h1, x, y1+h1]
            else:
                if e2.cy < self.cy:
                    # 2
                    # a 1
                    return [x, y, x2+w2, y, x2+w2, y1, x+w, y1, x+w, y+h, x, y+h]
                else:
                    # a 1
                    # 2
                    return [x, y, x+w, y, x+w, y1+h1, x2+w2, y1+h1, x2+w2, y+h, x, y+h]
        elif e2_same_row and e1_same_col:
            if e2.cx < self.cx:
                if e1.cy < self.cy:
                    #   1
                    # 2 a
                    return [x, y2, x1, y2, x1, y, x+w, y, x+w, y+h, x, y+h]
                else:
                    # 2 a
                    #   1
                    return [x, y, x+w, y, x+w, y+h, x1, y+h, x1, y2+h2, x, y2+h2]
            else:
                if e1.cy < self.cy:
                    # 1
                    # a 2
                    return [x, y, x1+w1, y, x1+w1, y2, x+w, y2, x+w, y+h, x, y+h]
                else:
                    # a 2
                    # 1
                    return [x, y, x+w, y, x+w, y2+h2, x1+w1, y2+h2, x1+w1, y+h, x, y+h]
        else:
            return []

def points_to_rounded_path(points, r):
    result = []
    points.extend(points[:6])
    for i in range(0, len(points) - 6, 2):
        (x1, y1, x2, y2, x3, y3) = points[i:i+6]
        if x1 < x2 and y2 > y3:
            #   3
            # 1 2
            (x2, c, dx, dy) = (x2 - r, 0, r, -r)
        elif x1 < x2 and y2 < y3:
            # 1 2
            #   3
            (x2, c, dx, dy) = (x2 - r, 1, r, r)
        elif x1 > x2 and y2 < y3:
            # 2 1
            # 3
            (x2, c, dx, dy) = (x2 + r, 0, -r, r)
        elif x1 > x2 and y2 > y3:
            # 3
            # 2 1
            (x2, c, dx, dy) = (x2 + r, 1, -r, -r)
        elif y1 < y2 and x2 > x3:
            #   1
            # 3 2
            (y2, c, dx, dy) = (y2 - r, 1, -r, r)
        elif y1 < y2 and x2 < x3:
            # 1
            # 2 3
            (y2, c, dx, dy) = (y2 - r, 0, r, r)
        elif y1 > y2 and x2 < x3:
            # 2 3
            # 1
            (y2, c, dx, dy) = (y2 + r, 1, r, -r)
        elif y1 > y2 and x2 > x3:
            # 3 2
            #   1
            (y2, c, dx, dy) = (y2 + r, 0, -r, -r)
        result.append(f"L{x2} {y2} a{r} {r} 0 0 {c} {dx} {dy}")
    result[0:0] = [f"M{x2+dx} {y2+dy}"]
    return " ".join(result)



irregular_plurals = {
 'aieul': 'aieux',
 'aval': 'avals',
 'bail': 'baux',
 'bal': 'bals',
 'betail': 'bestiaux',
 'bijou': 'bijoux',
 'bleu': 'bleus',
 'caillou': 'cailloux',
 'cal': 'cals',
 'carnaval': 'carnavals',
 'chacal': 'chacals',
 'choral': 'chorals',
 'chou': 'choux',
 'ciel': 'cieux',
 'corail': 'coraux',
 'credit-bail': 'credits-baux',
 'email': 'emaux',
 'emeu': 'emeus',
 'emmenthal': 'emmenthals',
 'enfeu': 'enfeus',
 'etal': 'etals',
 'fatal': 'fatals',
 'festival': 'festivals',
 'gemmail': 'gemmaux',
 'genou': 'genoux',
 'gentilhomme': 'gentilshommes',
 'glacial': 'glacials',
 'hibou': 'hiboux',
 'joujou': 'joujoux',
 'landau': 'landaus',
 'madame': 'mesdames',
 'mademoiselle': 'mesdemoiselles',
 'mistral': 'mistrals',
 'monsieur': 'messieurs',
 'natal': 'natals',
 'naval': 'navals',
 'oeil': 'yeux',
 'oeil-de-boeuf': 'oeils-de-boeuf',
 'oeil-de-chat': 'oeils-de-chat',
 'pal': 'pals',
 'pascal': 'pascals',
 'perinatal': 'perinatals',
 'pneu': 'pneus',
 'postnatal': 'postnatals',
 'pou': 'poux',
 'prenatal': 'prenatals',
 'recital': 'recitals',
 'regal': 'regals',
 'sarrau': 'sarraus',
 'soupirail': 'soupiraux',
 'travail': 'travaux',
 'unau': 'unaus',
 'vantail': 'vantaux',
 'veto': 'veto',
 'vitrail': 'vitraux'
}

def pluralize(word):
    word = word.lower()
    if word in irregular_plurals:
        return irregular_plurals[word]
    if word.endswith(("eau", "oeu")):
        return word + "x"
    if word.endswith(("s", "x", "z")):
        return word
    if word.endswith("al"):
        return word[:-2] + "aux"
    if word.endswith(("au", "eu")):
        return word + "x"
    return word + "s"



from ast import literal_eval
import re

def parse_value(v):
    v = v.strip()
    if v.startswith("'"):
        if v != "''":
            v = v.replace("''", "\\'")
        v = literal_eval(f"r{v}")
        v = v.replace("\\'", "'")
        v = re.sub(r"(?<!\\)\\n", "\n", v) # replace \n into newline, except if preceded by \
        v = v.replace(r"\\n", r"\n") # replace stuff like \\newcommand into \newcommand
        v = re.sub(r"(?<!\\)\\t", "\t", v) # idem for \t
        v = v.replace(r"\\t", r"\t") # idem for \t
        v = v.replace("\\xa0", '\xa0') # replace \xa0 into non-breaking space
        return v
    elif v.lower() == "true":
        return True
    elif v.lower() == "false":
        return False
    elif v.lower() == "null":
        return None
    elif v:
        return literal_eval(v)
    else:
        return []

def run(path):
    text = path.read_text(encoding="utf8")
    data = {}
    lines = text.splitlines()
    i = 0
    while i < len(lines):
        line = lines[i]
        i += 1
        (k1, v1) = re.match(r"(\S+?) *: *(.*?) *$", line).groups()
        data[k1] = parse_value(v1)
        if data[k1] == []: # list of dicts
            while i < len(lines):
                line = lines[i]
                (prefix, k2, v2) = re.match(r"([- ]*)(\w+?) *: *(.*?) *$", line).groups()
                if prefix == "  - ": # new dict
                    data[k1].append({k2: parse_value(v2)})
                    i += 1
                elif prefix == "    ": # continuation of previous dict
                    data[k1][-1][k2] = parse_value(v2)
                    i += 1
                else:
                    break
    return data



from ..parse_mcd import Visitor
from ..tools.parser_tools import first_child, parse_source, reconstruct_source


class Numbering(Visitor):

    def __init__(self, tree, params):

        self.size = len(str(len(set(node.children[0].value for node in tree.find_data("box_name")))))

        def new_name(base_name, name, i):
            if name == params["df"]:
                return name
            return f"{base_name} {i:0{self.size}d}_"

        def update_translations(token_name, base_name, n):
            nodes = tree.find_data(token_name)
            names = [node.children[0].children[0].value for node in nodes]
            self.names.update({name: new_name(base_name, name, i) for (i, name) in enumerate(names, n)})
            self.numbers.update({new_name(base_name, name, i): f"{i:0{self.size}d}"  for (i, name) in enumerate(names, n)})
        
        self.names = {}
        self.numbers = {}
        update_translations("entity_name_def", _("ENTITY"), 1)
        update_translations("assoc_name_def", _("ASSOC"), len(self.names) + 1)

        self.attr_base = _("attr")
        self.leg_note_base = _("role")

    def update_box_attrs(self, tree, token_name):
        number = self.numbers[first_child(tree, token_name).children[0].value]
        for (i, attr) in enumerate(tree.find_data("attr"), 1):
            if attr.children[0] == "":
                continue # do nothing for a spacer attribute
            attr.children[0].value = f"{self.attr_base} {number} {i}"
    
    def box_name(self, tree):
        token = tree.children[0]
        token.value = self.names[token.value]

    def entity_clause(self, tree):
        self.update_box_attrs(tree, "entity_name_def")
    
    def assoc_clause(self, tree):
        self.update_box_attrs(tree, "assoc_name_def")
        for (i, attr) in enumerate(tree.find_data("leg_note"), 1):
            attr.children[0].value = f"{self.leg_note_base} {i}"


def run(source, params, **kargs):
    tree = parse_source(source)
    visitor = Numbering(tree, params)
    visitor.visit(tree)
    return reconstruct_source(tree)





from collections import Counter
import contextlib
from importlib import import_module
import re
from pathlib import Path

from .tools.parser_tools import parse_source
from .tools.string_tools import rstrip_digit_or_underline

BLACKLIST = {
    "fr": ["date", "calendrier", "période"],
    "en": ["date", "calendar", "period"]
}

def guess_title(source, language):
    """
    Find the name of the most referenced entity in the MCD
    and pluralize it.
    """
    blacklist = BLACKLIST.get(language, [])
    tree = parse_source(source)
    names = []
    for node in tree.find_data("entity_name_ref"):
        name = rstrip_digit_or_underline(node.children[0].children[0])
        if name.lower() not in blacklist:
            names.append(name)
    counter = Counter(names)
    if not counter:
        return ""
    title = counter.most_common(1)[0][0]
    title = re.sub(r"[^\w '\._-]", "-", title)
    with contextlib.suppress(ModuleNotFoundError):
        pluralize = import_module(f"tools.pluralize_{language}").pluralize
        title = " ".join(map(pluralize, title.split()))
    return title.capitalize()


def may_update_params_with_guessed_title(source, params):
    if not params.get("guess_title"):
        return
    language = params["language"][:2]
    title = guess_title(source, language)
    if not title:
        return
    Path(f"{params['output_name']}_new_title.txt").write_text(title, encoding="utf8")
    params["title"] = title
    params["output_name"] = str(Path(params["output_dir"], title))



import re

from .attribute import *
from .leg import ConstraintLeg
from .tools.string_tools import is_a_description, raw_to_bid

class Constraint:

    counter = 0

    @classmethod
    def reset_counter(cls):
        cls.counter = 0

    def __init__(self, clause):
        self.source = clause["source"]
        self.raw_name = clause.get("name", "Anonymous")
        self.name_view = clause.get("name", "")
        Constraint.counter += 1
        self.bid = f'{raw_to_bid(self.raw_name)}_CONSTRAINT_#{Constraint.counter}'
        self.note = clause.get("constraint_note")
        self.legs = []
        for target in clause.get("constraint_targets", []):
            self.legs.append(ConstraintLeg(self, target.get("constraint_leg", ""), target["name"]))
        if self.legs:
            self.coords = clause.get("constraint_coords", [])
        else:
            # When there is neither coord nor target, the constraint is placed at the top right corner
            # which is obviously undesirable, forcing the user to specify at least one coord.
            self.coords = clause.get("constraint_coords", [0, 0])
        self.kind= "constraint"

    def register_boxes(self, boxes):
        self.boxes = boxes

    def calculate_size(self, style, get_font_metrics):
        constraint_font = get_font_metrics(style["constraint_font"])
        self.get_constraint_string_width = constraint_font.get_pixel_width
        self.constraint_height = constraint_font.get_pixel_height()
        if self.name_view.strip():
            size = max(self.get_constraint_string_width(self.name_view), self.constraint_height)
        else:
            size = self.get_constraint_string_width("W") * len(self.name_view)
        self.w = self.h = style["constraint_margin"] * 2 + size

    def register_center(self, geo):
        self.page_width = geo["width"]
        self.page_height = geo["height"]
        if self.coords:
            # The center of a constraint is either at a certain ratio of the width and height of the
            # page, or aligned with the center of a box.
            if isinstance(self.coords[0], (float, int)):
                self.cx = self.coords[0] * self.page_width // 100
            else:
                self.cx = geo["cx"][raw_to_bid(self.coords[0])]
            if isinstance(self.coords[1], (float, int)):
                self.cy = self.coords[1] * self.page_height // 100
            else:
                self.cy = geo["cy"][raw_to_bid(self.coords[1])]
        else:
            # The center of a constraint is the barycenter of the centers of its boxes
            self.cx = sum(geo["cx"][leg.bid] for leg in self.legs) / len(self.legs)
            self.cy = sum(geo["cy"][leg.bid] for leg in self.legs) / len(self.legs)
        self.l = self.cx - self.w // 2
        self.r = self.cx + self.w // 2
        self.t = self.cy - self.h // 2
        self.b = self.cy + self.h // 2

    def _description(self, style):
        return [
            (
                "circle_with_note" if is_a_description(self.note) else "circle",
                {
                    "stroke_depth": style["constraint_stroke_depth"],
                    "stroke_color": style["constraint_stroke_color"],
                    "color": style["constraint_background_color"],
                    "cx": self.cx,
                    "cy": self.cy,
                    "r": self.w // 2,
                    "note": self.note,
                },
            ),
            (
                "text_above_note",
                {
                    "text": self.name_view,
                    "text_color": style['constraint_text_color'],
                    "x": self.cx - self.get_constraint_string_width(self.name_view) / 2,
                    "y": self.cy + self.constraint_height * style["constraint_text_height_tweak"],
                    "family": style["card_font"]["family"],
                    "size": style["card_font"]["size"],
                },
            ),
        ]

    def description(self, style, geo):
        result = []
        result.append(("comment", {"text": f"Constraint {self.bid}"}))
        result.append(
            (
                "begin_component",
                {
                    "page": self.page,
                    "visibility": "hidden" if self.page else "visible",
                }
            )
        )
        result.extend(self.leg_descriptions(style, geo))
        result.append(("begin_group", {}))
        result.extend(self._description(style))
        result.append(("end", {}))
        result.append(("end", {}))
        return result

    def leg_descriptions(self, style, geo):
        result = []
        for leg in self.legs:
            result.extend(leg.description(style, geo))
        return result

    def invert_coords_horizontal_mirror(self):
        if self.coords and isinstance(self.coords[1], (float, int)):
            self.coords[1] = 100 - self.coords[1]
            self.source = re.sub(r"(.+):.+", fr"\1: {self.coords[0]}, {self.coords[1]}", self.source)
    
    def invert_coords_vertical_mirror(self):
        if self.coords and isinstance(self.coords[0], (float, int)):
            self.coords[0] = 100 - self.coords[0]
            self.source = re.sub(r"(.+):.+", fr"\1: {self.coords[0]}, {self.coords[1]}", self.source)
    
    def invert_coords_diagonal_mirror(self):
        if self.coords:
            (self.coords[0], self.coords[1]) = (self.coords[1], self.coords[0])
            self.source = re.sub(r"(.+):.+", fr"\1: {self.coords[0]}, {self.coords[1]}", self.source)



import itertools
from collections import defaultdict
from hashlib import md5
import json
from pathlib import Path

from .association import Association
from .attribute import Attribute
from .constraint import Constraint
from .diagram_link import DiagramLink
from .entity import Entity
from .grid import Grid
from .inheritance import Inheritance
from .mocodo_error import MocodoError
from .phantom import Phantom
from .tools.string_tools import raw_to_bid
from .tools.parser_tools import extract_clauses

def cmp(x, y):
    return (x > y) - (x < y)

SYS_MAXINT = 9223372036854775807 # an integer larger than any practical list or string index

class Mcd:

    def __init__(self, source, get_font_metrics=None, **params):
        
        def calculate_uid():
            h = md5(source.encode("utf-8")).hexdigest()
            if params.get("uid_suffix"):
                return f"{h[:8]}_{params['uid_suffix']}"
            else:
                return h[:8]

        def create():
            self.entities = {}
            self.associations = {}
            self.constraints = []
            self.inheritances = []
            mcd_has_cif = False
            self.commented_lines = []
            self.constraint_clauses = []
            seen = set()
            self.rows = [[]]
            pages = defaultdict(list)
            for clause in extract_clauses(source):
                indentation = clause.get("indent", "")
                if clause["type"] == "break":
                    self.rows.append([])
                    continue
                if clause["type"] == "comment":
                    if not self.rows[-1]:
                        self.commented_lines.append(clause["text"])
                    continue
                if clause["type"] == "phantoms":
                    phantoms = [Phantom() for _ in range(clause["count"])]
                    if self.rows[-1]:
                        self.rows[-1].extend(phantoms)
                    else:
                        self.rows.append(phantoms)
                    continue
                if clause["type"] == "constraint":
                    element = Constraint(clause)
                    if element.name_view == "CIF":
                        mcd_has_cif = True
                    self.constraints.append(element)
                    pages[indentation].append(element)
                    self.constraint_clauses.append(clause)
                    continue
                if clause["type"] == "inheritance":
                    element = Inheritance(clause, **params)
                    self.inheritances.append(element)
                    pages[indentation].append(element)
                else:
                    if clause["type"] == "association":
                        element = Association(clause, **params)
                        if element.bid in self.associations:
                            raise MocodoError(7, _('Duplicate association "{name}". If you want to make two associations appear with the same name, you must suffix it with a number.').format(name=element.raw_name)) # fmt: skip
                        self.associations[element.bid] = element
                        pages[indentation].append(element)
                    elif clause["type"] == "entity":
                        element = Entity(clause)
                        if element.bid in self.entities:
                            raise MocodoError(6, _('Duplicate entity "{name}". If you want to make two entities appear with the same name, you must suffix it with a number.').format(name=element.raw_name)) # fmt: skip
                        self.entities[element.bid] = element
                        pages[indentation].append(element)
                    else:
                        raise NotImplementedError
                    if element.bid in seen:
                        raise MocodoError(8, _('One entity and one association share the same name "{name}".').format(name=element.raw_name)) # fmt: skip
                    seen.add(element.bid)
                self.rows[-1].append(element)
            if not seen:
                raise MocodoError(4, _('The ERD "{title}" is empty.').format(title=params["title"])) # fmt: skip
            self.rows = [row for row in self.rows if row]
            self.col_count = max(len(row) for row in self.rows)
            self.row_count = len(self.rows)
            self.page_count = len(pages)
            for (i, (indentation, elements)) in enumerate(sorted(pages.items(), key=lambda x: x[0])):
                for element in elements:
                    element.page = i
            self.header = "\n".join(self.commented_lines) + "\n\n" if self.commented_lines else ""
            for association in self.associations.values():
                association.register_mcd_has_cif(mcd_has_cif)
        
        def add_legs():
            for association in self.associations.values():
                for leg in association.legs:
                    if leg.entity_bid in self.entities:
                        entity = self.entities[leg.entity_bid]
                    elif leg.entity_bid in self.associations:
                        raise MocodoError(18, _('Association "{association}" linked to another association "{entity}"!').format(association=association.bid, entity=leg.entity_bid)) # fmt: skip
                    else:
                        raise MocodoError(1, _('Association "{association}" linked to an unknown entity "{entity}"!').format(association=association.bid, entity=leg.entity_bid)) # fmt: skip
                    leg.register_entity(entity)
            for inheritance in self.inheritances:
                for leg in inheritance.legs:
                    if leg.entity_bid in self.entities:
                        entity = self.entities[leg.entity_bid]
                    elif leg.entity_bid in self.associations:
                        raise MocodoError(44, _('Inheritance "{inheritance}" linked to an association "{entity}"!').format(inheritance=inheritance.bid, entity=leg.entity_bid))
                    else:
                        raise MocodoError(42, _('Inheritance "{inheritance}" linked to an unknown entity "{entity}"!').format(inheritance=inheritance.bid, entity=leg.entity_bid))
                    leg.register_entity(entity)
            for constraint in self.constraints:
                for leg in constraint.legs:
                    if leg.bid in self.associations:
                        box = self.associations[leg.bid]
                    elif leg.bid in self.entities:
                        box = self.entities[leg.bid]
                    else:
                        raise MocodoError(40, _('Constraint "{constraint}" linked to an unknown entity or association "{box}"!').format(constraint=constraint.bid, box=leg.bid)) # fmt: skip
                    leg.register_box(box)
                for coord in constraint.coords:
                    if isinstance(coord, (float, int)):
                        continue
                    bid = raw_to_bid(coord)
                    if bid in self.associations or bid in self.entities:
                        continue
                    raise MocodoError(43, _('Constraint "{constraint}" aligned with an unknown entity or association "{box}"!').format(constraint=constraint.bid, box=bid)) # fmt: skip
        
        def add_attributes():
            strengthening_legs = dict((entity_bid, []) for entity_bid in self.entities)
            for association in self.associations.values():
                for leg in association.legs:
                    if leg.kind == "strengthening":
                        strengthening_legs[leg.entity_bid].append(leg)
            children = set()
            for inheritance in self.inheritances:
                for leg in inheritance.legs[1:]: # the first leg is the parent
                    children.add(leg.entity_bid) # the other legs are its children
            Attribute.id_gutter_strong_string = params["id_gutter_strong_string"]
            Attribute.id_gutter_weak_string = params["id_gutter_weak_string"]
            Attribute.id_gutter_alts = params["id_gutter_alts"]
            for (entity_bid, entity) in self.entities.items():
                entity.add_attributes(
                    legs_to_strengthen=strengthening_legs[entity_bid],
                    is_child=entity_bid in children,
                    fk_format=params.get("fk_format", "#{label}")
                )
            self.has_alt_identifier = any(entity.has_alt_identifier for entity in self.entities.values())

        def check_weak_entities_without_discriminator():
            too_weak_entities = defaultdict(int)
            for association in self.associations.values():
                for leg in association.legs:
                    if leg.kind != "strengthening":
                        continue
                    for attribute in leg.entity.attributes:
                        if attribute.kind == "weak":
                            break
                    else: # the weak entity has no discriminator.
                        # Ensure the max cards of the other legs are 1
                        for other_leg in association.legs:
                            if other_leg is leg:
                                continue
                            if other_leg.card[1] != "1":
                                # Otherwise, accumulate them
                                too_weak_entities[leg.entity.bid] += 1
            for (too_weak_entity, count) in too_weak_entities.items():
                if count < 2: # one isolated "too weak entity"
                    raise MocodoError(50, _('The weak entity "{entity}" should have a discriminator.').format(entity=too_weak_entity)) # fmt: skip
        
        def set_id_gutter_visibility():
            flag = params["id_gutter_visibility"]
            is_visible = flag == "on" or (flag == "auto" and self.has_alt_identifier)
            for entity in self.entities.values():
                entity.set_id_gutter_visibility(is_visible)
        
        def tweak_straight_cards():
            coordinates = {}
            for (j, row) in enumerate(self.rows):
                for (i, box) in enumerate(row):
                    coordinates[box] = (i, j)
            d = defaultdict(list)
            tweakable_legs = {}
            for association in self.associations.values():
                if association.is_invisible:
                    continue
                for leg in association.legs:
                    (ei, ej) = coordinates[leg.entity]
                    (ai, aj) = coordinates[leg.association]
                    vector = (cmp(ai, ei), cmp(aj, ej))
                    vector = (" SN"[cmp(aj, ej)] + " EW"[cmp(ai, ei)]).strip()
                    d[leg.entity].append(vector)
                    tweakable_legs[(leg.entity, vector)] = leg
            flex = params.get("flex", 0)
            for (entity, vectors) in d.items():
                for vector in vectors:
                    leg = tweakable_legs[(entity, vector)]
                    if not leg.card_view.strip():
                        continue
                    elif vector == "E":
                        if vectors.count("E") == 1 and "SE" in vectors and "NE" not in vectors:
                            leg.twist = True
                    elif vector == "S":
                        if vectors.count("S") == 1 and "SE" in vectors and "SW" not in vectors:
                            leg.twist = True
                    elif vector == "W":
                        if vectors.count("W") == 1 and "SW" in vectors and "NW" not in vectors:
                            leg.twist = True
                    elif vector == "N":
                        if vectors.count("N") == 1 and "NE" in vectors and "NW" not in vectors:
                            leg.twist = True
                    elif flex == 0:
                        continue
                    elif vector == "SE" and vectors.count("SE") == 1:
                        if vectors.count("E") > 1:
                            leg.set_spin_strategy(flex)
                        elif vectors.count("S") > 1:
                            leg.set_spin_strategy(-flex)
                    elif vector == "SW" and vectors.count("SW") == 1:
                        if vectors.count("S") > 1:
                            leg.set_spin_strategy(flex)
                        elif vectors.count("W") > 1:
                            leg.set_spin_strategy(-flex)
                    elif vector == "NW" and vectors.count("NW") == 1:
                        if vectors.count("W") > 1:
                            leg.set_spin_strategy(flex)
                        elif vectors.count("N") > 1:
                            leg.set_spin_strategy(-flex)
                    elif vector == "NE" and vectors.count("NE") == 1:
                        if vectors.count("N") > 1:
                            leg.set_spin_strategy(flex)
                        elif vectors.count("E") > 1:
                            leg.set_spin_strategy(-flex)
        
        def add_diagram_links():
            self.diagram_links = []
            for entity in self.entities.values():
                for attribute in entity.attributes:
                    if attribute.primary_entity_bid:
                        self.diagram_links.append(DiagramLink(self.entities, entity, attribute))
            
        def may_center():
            for row in self.rows:
                n = self.col_count - len(row)
                if n:
                    row[0:0] = [Phantom() for i in range(n // 2)]
                    row.extend(Phantom() for i in range(n // 2 + n % 2))
        
        def make_boxes():
            i = itertools.count()
            self.boxes = []
            for row in self.rows:
                for box in row:
                    box.index = next(i)
                    self.boxes.append(box)
                    box.register_boxes(self.boxes)
            self.box_count = len(self.boxes)

        # The following keys are actually created by __main__.py.
        # Using `get` instead of `[]` is for testing purposes only.
        params.setdefault("id_gutter_strong_string", "ID")
        params.setdefault("id_gutter_weak_string", "id")
        params.setdefault("id_gutter_alts", dict(zip("123456789", "123456789")))
        params.setdefault("id_gutter_visibility", "auto")

        self.get_font_metrics = get_font_metrics
        Phantom.reset_counter()
        Association.reset_df_counter()
        Inheritance.reset_counter()
        Constraint.reset_counter()
        self.uid = calculate_uid()
        create()
        self.update_footer()
        add_legs()
        add_attributes()
        check_weak_entities_without_discriminator()
        set_id_gutter_visibility()
        add_diagram_links()
        may_center()
        make_boxes()
        tweak_straight_cards()
        self.title = params.get("title", "Untitled")
    
    def update_footer(self):
        constraint_sources = [constraint.source for constraint in self.constraints]
        self.footer = "\n\n" + "\n".join(constraint_sources) if constraint_sources else ""

    def get_layout_data(self):
        successors = [set() for i in range(self.box_count)] # use `set` to deduplicate reflexive associations
        multiplicity = defaultdict(int) # but count the multiplicity (1 or 2) of each link
        for inheritance in self.inheritances:
            for leg in inheritance.legs:
                successors[inheritance.index].add(leg.entity.index)
                successors[leg.entity.index].add(inheritance.index)
                multiplicity[(inheritance.index, leg.entity.index)] += 1
                multiplicity[(leg.entity.index, inheritance.index)] += 1
        if self.associations:
            for association in self.associations.values():
                for leg in association.legs:
                    successors[association.index].add(leg.entity.index)
                    successors[leg.entity.index].add(association.index)
                    multiplicity[(association.index, leg.entity.index)] += 1
                    multiplicity[(leg.entity.index, association.index)] += 1
        else:
            for diagram_link in self.diagram_links:
                fei = diagram_link.foreign_entity.index
                pei = diagram_link.primary_entity.index
                if fei != pei:
                    successors[fei].add(pei)
                    successors[pei].add(fei)
                    multiplicity[(fei, pei)] += 1
                    multiplicity[(pei, fei)] += 1
        links = tuple((node, child) for (node, children) in enumerate(successors) for child in children if node < child)
        return {
            "links": links,
            "successors": successors,
            "col_count": self.col_count,
            "row_count": self.row_count,
            "multiplicity": dict(multiplicity)
        }
    
    def get_layout(self):
        return [box.index for row in self.rows for box in row]
    
    def get_row_text(self, row):
        return "\n".join(box.source for box in row)
    
    def get_non_phantom_count(self):
        return sum(box.kind != "phantom" for box in self.boxes)

    def set_layout(self, layout, col_count=None, row_count=None, **kwargs):
        if col_count and row_count:
            (self.col_count, self.row_count) = (col_count, row_count)
        def get_or_create_box(index):
            return Phantom() if layout[index] is None else self.boxes[layout[index]]
        i = itertools.count()
        self.rows = [[get_or_create_box(next(i)) for x in range(self.col_count)] for y in range(self.row_count)]
        def suppress_empty_rows(y):
            while self.rows: # there's at least one row
                for box in self.rows[y]:
                    if box.kind != "phantom":
                        return
                del self.rows[y]
                self.row_count -= 1
        suppress_empty_rows(0)
        suppress_empty_rows(-1)
        def suppress_empty_cols(x):
            while self.rows[0]: # there's at least one column
                for row in self.rows:
                    if row[x].kind != "phantom":
                        return
                for row in self.rows:
                    del row[x]
                self.col_count -= 1
        suppress_empty_cols(0)
        suppress_empty_cols(-1)
    
    def get_clauses(self):
        result = self.header
        if self.associations:
            result += "\n\n".join(self.get_row_text(row) for row in self.rows)
        else:
            result += "\n\n".join(":\n" + "\n:\n".join(self.get_row_text(row).split("\n")) + "\n:" for row in self.rows)
        return result + self.footer

    def get_vertically_flipped_clauses(self):
        for constraint in self.constraints:
            constraint.invert_coords_horizontal_mirror()
        self.update_footer()
        return self.header + "\n\n".join(self.get_row_text(row) for row in self.rows[::-1]) + self.footer
    
    def get_horizontally_flipped_clauses(self):
        for constraint in self.constraints:
            constraint.invert_coords_vertical_mirror()
        self.update_footer()
        return self.header + "\n\n".join(self.get_row_text(row[::-1]) for row in self.rows) + self.footer
    
    def get_diagonally_flipped_clauses(self):
        for constraint in self.constraints:
            constraint.invert_coords_diagonal_mirror()
        self.update_footer()
        return self.header + "\n\n".join(self.get_row_text(row) for row in zip(*self.rows)) + self.footer
    
    def get_refitted_clauses(self, nth_fit_or_col_count, row_count=None):
        if row_count:
            col_count = nth_fit_or_col_count
        else:
            nth_fit = nth_fit_or_col_count
            grid = Grid(len(self.boxes) + 100) # make sure there are enough precalculated grids
            start = len(self.entities) + len(self.associations) # number of nonempty boxes
            (col_count, row_count) = grid.get_nth_next(start, nth_fit)
        result = []
        i = 0
        for box in self.boxes:
            if box.kind != "phantom":
                if i % col_count == 0 and i:
                    result.append("")
                result.append("  " * box.page + box.source.rstrip())
                i += 1
        for i in range(i, col_count * row_count):
            if i % col_count == 0 and i:
                result.append("")
            result.append(":")
        return self.header + "\n".join(result) + self.footer


    def calculate_or_retrieve_geo(self, params):
        geo_path = Path(f"{params['output_name']}_geo.json")
        if geo_path.is_file() and params["scale"] == 1 and params["reuse_geo"]:
            try:
                web_geo = json.loads(geo_path.read_text(encoding="utf8"))
            except:
                raise MocodoError(33, _('Unable to reuse the geometry file "{filename}".').format(filename=geo_path)) # fmt: skip
            # convert lists of couples to dicts
            geo = {}
            for (k, v) in web_geo.items():
                if isinstance(v, list):
                    geo[k] = dict(v)
                else:
                    geo[k] = v
            return geo
        geo = {
            "width": self.w,
            "height": self.h,
            "cx": {
                box.bid: box.x + box.w // 2
                for row in self.rows
                for box in row
                if box.kind != "phantom"
            },
            "cy": {
                box.bid: box.y + box.h // 2
                for row in self.rows
                for box in row
                if box.kind != "phantom"
            },
            "shift": {
                leg.lid: 0
                for row in self.rows
                for box in row
                for leg in box.legs
                if hasattr(leg, "card_view")},
            "ratio": {
                leg.lid: 1.0
                for row in self.rows
                for box in row
                for leg in box.legs
                if leg.arrow
            },
        }
        web_geo = {k: list(v.items()) if isinstance(v, dict) else v for (k, v) in geo.items()}
        text = json.dumps(web_geo, indent=2, ensure_ascii=False)
        text = text.replace("\n      ", " ")
        text = text.replace("\n    ]", " ]")
        text = text + "\n"
        try:
            geo_path.write_text(text, encoding="utf8")
        except IOError:
            raise MocodoError(34, _('Unable to save geometry file "{filename}".').format(filename=geo_path)) # fmt: skip
        return geo


    def calculate_size(self, style):

        def increase_margins_in_presence_of_clusters():
            if not self.associations: # relational diagram or MCD without associations
                return
            factor = 1 + max(a.peg_count for a in self.associations.values())
            style["margin"] *= factor
            style["card_margin"] *= factor

        def card_max_width():
            get_pixel_width = self.get_font_metrics(style["card_font"]).get_pixel_width
            cardinalities = {"0,N"} # default value, in case there is no cardinalities at all
            for association in self.associations.values():
                for leg in association.legs:
                    cardinalities.add(leg.card_view.strip("_"))
            return max(map(get_pixel_width, cardinalities))
        #
        def calculate_sizes():
            for row in self.rows:
                for (i, box) in enumerate(row):
                    box.calculate_size(style, self.get_font_metrics)
                    max_box_width_per_column[i] = max(box.w, max_box_width_per_column[i])
            for diagram_link in self.diagram_links:
                diagram_link.calculate_size(style, self.get_font_metrics)
            for constraint in self.constraints:
                constraint.calculate_size(style, self.get_font_metrics)
        #
        def make_horizontal_layout():
            self.w = style["margin"]
            for row in self.rows:
                horizontal_offset = style["margin"]
                for (i, box) in enumerate(row):
                    box.x = horizontal_offset + (max_box_width_per_column[i] - box.w) // 2
                    horizontal_offset += max_box_width_per_column[i] + join_width
                self.w = max(self.w, horizontal_offset)
            self.w += style["margin"] - join_width
        #
        def compress_horizontally():
            dx = 0
            for i in range(1, self.col_count):
                dx = SYS_MAXINT
                for row in self.rows:
                    b1 = row[i-1]
                    b2 = row[i]
                    space = b2.x - b1.x - b1.w - join_width
                    dx = min(dx, space)
                for row in self.rows:
                    row[i].x -= dx
            self.w -= dx
        #
        def make_vertical_layout():
            vertical_offset = style["margin"]
            for row in self.rows:
                max_box_height = max(box.h for box in row)
                for box in row:
                    box.y = vertical_offset + (max_box_height - box.h) // 2
                vertical_offset += max_box_height + join_height
            self.h = vertical_offset + style["margin"] - join_height
        #
        def compress_vertically():
            dy = 0
            for j in range(1, self.row_count):
                dy = SYS_MAXINT
                for (i2, b2) in enumerate(self.rows[j]):
                    y1_max = 0
                    for (i1, b1) in enumerate(self.rows[j-1]):
                        if (i1 == i2) or (b1.x < b2.x < b1.x + b1.w + join_width) or (b1.x - join_width < b2.x + b2.w < b1.x + b1.w):
                            y1_max = max(y1_max, b1.y + b1.h)
                    space = b2.y - y1_max - join_height
                    dy = min(dy, space)
                for box in self.rows[j]:
                    box.y -= dy
            self.h -= dy
        #

        increase_margins_in_presence_of_clusters()
        style["card_max_width"] = card_max_width()
        style["card_max_height"] = self.get_font_metrics(style["card_font"]).get_pixel_height()
        join_width  = 2 * style["card_margin"] + style["card_max_width"]
        join_height = 2 * style["card_margin"] + style["card_max_height"]
        max_box_width_per_column = [0] * self.col_count
        calculate_sizes()
        make_horizontal_layout()
        compress_horizontally()
        make_vertical_layout()
        compress_vertically()
    
    def description(self, style, geo):
        for box in self.boxes:
            box.register_center(geo)
        for constraint in self.constraints:
            constraint.register_center(geo)
        result = []
        for element in self.entities.values():
            result.extend(element.description(style, geo))
        for element in self.associations.values():
            result.extend(element.description(style, geo))
        for element in self.inheritances:
            result.extend(element.description(style, geo))
        for element in self.diagram_links:
            result.extend(element.description(style, geo))
        for element in self.constraints:
            result.extend(element.description(style, geo))
        result.append(("comment", {"text": "Notes"}))
        result.append(
            (
                "notes",
                {
                    "height_threshold": geo["height"] - style["note_overlay_height"] - style["card_margin"],
                    "overlay_height": style["note_overlay_height"],
                    "x": geo["width"] // 2,
                    "y_top": style["note_baseline"],
                    "y_bottom": geo["height"] - style["note_overlay_height"] + style["note_baseline"],
                    "y": geo["height"] - style["note_overlay_height"],
                    "color": style["note_color"],
                    "text_color": style["note_text_color"],
                    "opacity": style["note_opacity"],
                    "font_family": style["note_font"]["family"],
                    "font_size": style["note_font"]["size"],
                }
            )
        )
        if self.page_count > 1:
            diameter = style["note_overlay_height"] / 4
            result.append(("comment", {"text": "Pager"}))
            for i in range(self.page_count):
                result.append(
                    (
                        "pager_dot",
                        {
                            "cx": geo["width"] / 2 - (self.page_count - 1 - 2 * i) * diameter,
                            "cy": geo["height"] + 2 * diameter,
                            "r": diameter / 2,
                            "color": "lightgray" if i else "gray",
                            "page": i,
                        }
                    )
                )
            result.append(("pager", {"page_count": self.page_count}))
        return result
    
    def get_overlaps(self):
        """
        Detect the cases when two legs overlap each other or when a leg overlaps a box.
        
        The detection is based on the assumption that the boxes are arranged in a grid
        (and doesn't take into account any tweaks that may have been applied to the box
        centers). To unify the detection, the boxes are treated as zero-length legs.
        Only horizontal and vertical overlaps are detected.
        """
        coordinates = {}
        for (j, row) in enumerate(self.rows):
            for (i, box) in enumerate(row):
                coordinates[box] = (i, j)
        segments = defaultdict(list)
        for entity in self.entities.values():
            if entity.is_invisible:
                continue
            (ie, je) = coordinates[entity]
            segments["i", ie].append((je, je, entity, entity))
            segments["j", je].append((ie, ie, entity, entity))
        for association in self.associations.values():
            if association.is_invisible:
                continue
            (ia, ja) = coordinates[association]
            segments[("i", ia)].append((ja, ja, association, association))
            segments[("j", ja)].append((ia, ia, association, association))
            for leg in association.legs:
                (ie, je) = coordinates[leg.entity]
                if ia == ie:
                    segments[("i", ia)].append((*sorted([ja, je]), association, leg.entity))
                elif ja == je:
                    segments[("j", ja)].append((*sorted([ia, ie]), association, leg.entity))
        result = []
        for quadruples in segments.values():
            quadruples.sort()
            for (l1, r1, e1, a1), (l2, r2, e2, a2) in zip(quadruples, quadruples[1:]):
                if e1.bid == e2.bid and a1.bid == a2.bid: # reflexive association
                    continue
                if l2 < r1:
                    result.append((e1.bid, a1.bid, e2.bid, a2.bid))
        return result

if __name__=="__main__":
    from .argument_parser import parsed_arguments
    source = """
        CLIENT: Réf. client, Nom, Prénom, Adresse
        PASSER, 0N CLIENT, 11 COMMANDE
        COMMANDE: Num commande, Date, Montant
        INCLURE, 1N COMMANDE, 0N PRODUIT: Quantité
        PRODUIT: Réf. produit, Libellé, Prix unitaire
    """.replace("  ", "").split("\n")
    params = parsed_arguments()
    mcd = Mcd(source, **params)
    print(mcd.get_horizontally_flipped_clauses())



import random
import re
from pathlib import Path

from .damerau_levenshtein import distance
from ..mocodo_error import MocodoError

MIN_DISTANCE = 3  # Minimal Damereau-Levenshtein distance between two words.

def random_words_generator(
    lorem_text,
    find_all_words=re.compile(r"(?u)[^\W\d_]{3,}").findall,
):
    # use dict.fromkeys instead of set for preserving order
    words = list(dict.fromkeys(word.lower() for word in find_all_words(lorem_text)))
    random.shuffle(words)
    previous_words = set()
    for word in words:
        for previous_word in previous_words:
            if distance(word, previous_word) < MIN_DISTANCE:
                # Keeping this word would be too confusing, let's try the next one.
                break
        else:
            yield word
            previous_words.add(word)


def obfuscator_factory(pool, params):

    pool = pool or ""

    # Initialize the random word generator
    try:
        lorem_path = Path(pool)
        lorem_text = lorem_path.read_text(encoding="utf8")
    except IOError:
        lorem_dir = Path(params["script_directory"], "resources", "lorem")
        lorem_path = lorem_dir / f"{lorem_path.stem}.txt"
        try:
            lorem_text = lorem_path.read_text(encoding="utf8")
        except IOError:
            lorem_path = lorem_dir / "lorem.txt"
            lorem_text = lorem_path.read_text(encoding="utf8")
    random_word = random_words_generator(lorem_text)

    # Define and return the inner obfuscator function
    cache = {}
    def obfuscate(name):
        suffix = ""
        if name not in cache or name == params["df"]:
            try:
                new_name = next(random_word)
            except StopIteration:
                raise MocodoError(12, _('Obfuscation failed. Not enough substitution words in "{filename}".').format(filename=lorem_path)) from None # fmt: skip
            if name.isupper():
                new_name = new_name.upper()
            elif name.islower():
                new_name = new_name.lower()
            elif name.partition(" ")[0].istitle():
                new_name = new_name.title()
            cache[name] = new_name
        return cache[name] + suffix
    return obfuscate



import contextlib
import random
import re

from ..tools.parser_tools import parse_source
from ..mocodo_error import MocodoError


def random_booleans(n, p):
    p = min(n, p)
    result = [True] * p + [False] * (n - p)
    random.shuffle(result)
    return result

def calculate_cards(scheme, refs):
    if len(refs) > len(set(refs)): # If the association is reflexive
        # Process the cases of n-ary associations with n > 2
        # (the other ones have already been filtered out by the caller)
        scheme = scheme.replace("_", "") # silently forbid weak associations
        scheme = scheme.replace("/", "") # silently forbid cluster associations
    (head, tail) = scheme.split("-")
    cards = [head] + [tail] * (len(refs) - 1)
    cards = [card.replace("*", random.choice("01")) for card in cards]
    random.shuffle(cards)
    return cards

def extract_weak_entities(cards, refs):
    for (card, ref) in zip(cards, refs):
        if "_" in card:
            yield ref

def biased_choice(elements):
    # Return a random element from a list, with a bias towards the last elements.
    i = max(random.randrange(0, len(elements)) for _ in range(3))
    return elements[i]

def biased_sample(elements, k):
    # Return a random sample of k elements from a list, with a bias towards the last elements.
    max_sample = [-1] # incorrect sample, but it will be replaced on the first iteration
    for _ in range(3):
        sample = random.sample(range(len(elements)), k)
        if min(sample) > min(max_sample):
            max_sample = sample
    return [elements[i] for i in max_sample]

def run(source, subargs=None, params=None, **kargs):
    match_card_scheme = re.compile(r"^[/_]?[01*][1N]-[/_]?[01*][1N]$").match
    subargs = subargs or {}
    default = {
            "n": 10,
            "arity_1": 2,
            "arity_3": 2,
            "arity_4": 0,
            "ent_attrs": 4,
            "doubles": 1,
            "composite_ids": 1,
            "assoc_attrs": 2,
            "*1-*N": 3,
            "01-11": 1,
            "from_scratch": False,
        }
    # Complete the given subargs with default values
    settings = {}
    for (k, v) in default.items():
        if k in subargs:
            cast = type(v) # cast to the same type as the default value
            with contextlib.suppress(TypeError):
                settings[k] = cast(subargs[k])
                continue
            # If the cast fails, silently fall back to the default value
        settings[k] = v
    
    # Complete the settings with the remaining card schemes
    for (k, v) in subargs.items():
        if match_card_scheme(k):
            with contextlib.suppress(TypeError):
                settings[k] = int(v)
    
    # Silently correct incompatible numbers
    n = settings["n"]
    for (k, v) in list(settings.items()):
        if k in ("n", "assoc_attrs"):
            continue
        elif k == "ent_attrs":
            settings[k] = max(v, 1)
        elif k in ("doubles", "arity_1", "arity_3", "arity_4"):
            settings[k] = min(v, n - 1)

    association_bases = [None, _("Reflexive"), _("Binary"), _("Ternary"), _("Quaternary")]
    entity_base = _("Entity")
    weak_entity_base = _("Weak Entity")
    attr_base = _("attr")
    id_base = _("id")

    card_schemes = []
    for (k, v) in settings.items():
        if match_card_scheme(k):
            card_schemes.extend([k] * v)
    card_schemes.extend(["*N-*N"] * (n - len(card_schemes)))

    # Try to find a combination of card schemes and arities that is compatible
    arities = [1] * settings["arity_1"] + [3] * settings["arity_3"] + [4] * settings["arity_4"] + [2] * n
    arities = arities[:n]
    max_tries = 100
    while max_tries > 0:
        random.shuffle(arities)
        # Slightly push the non binary arities towards the end (one iteration of bubble sort).
        # Reason: the non binary arities are tougher to place when there are not many entities yet
        for i in range(len(arities) - 1):
            if arities[i] != 2 and arities[i + 1] == 2:
                (arities[i], arities[i + 1]) = (arities[i + 1], arities[i])
        random.shuffle(card_schemes) # avoid being stuck with a card scheme incompatible with a small arity
        max_tries -= 1
        for (scheme, arity) in zip(card_schemes, arities):
            if "_" in scheme and arity == 1:
                break # A reflexive association cannot be weak
            if "/" in scheme and arity < 3:
                break # A cluster association is better with at least 3 entities
            if "11" in scheme and (arity > 2 or "/" in scheme):
                break # A DF is better with at most 2 entities or a cluster
        else:
            # All associations are compatible with their card schemes
            break
    else:
        raise MocodoError(28, _("Cannot find a suitable combination of card schemes and arities."))  # fmt: skip

    for key in ("assoc_attrs", "composite_ids"):
        settings[key] = random_booleans(n, settings[key])
    
    ent_attr_names = []
    for is_composite in settings["composite_ids"]:
        if is_composite:
            ent_attr_names.append([f"{id_base}", f"_{id_base}"] + [f"{attr_base}"] * random.randint(0, settings["ent_attrs"] - 2))
        else:
            ent_attr_names.append([f"{id_base}"] + [f"{attr_base}"] * random.randint(0, settings["ent_attrs"] - 1))
    
    if "from_scratch" in subargs:
        source = f"ENTITY_NAME_PLACEHOLDER 1_: {id_base} 1 1, {attr_base} 1 2, {attr_base} 1 3\n"

    tree = parse_source(source)
    entities = [node.children[0].children[0].value for node in tree.find_data("entity_name_def")]
    associations = [node.children[0].children[0].value for node in tree.find_data("assoc_name_def")]
    counter = len(entities) + len(associations) + 1

    clauses = [source]
    ref_pool = []
    weak_entities = set()
    for i in range(settings["n"] - settings["doubles"]):

        if arities[i] == 1:
            # to keep the MCD connected, don't create a new entity if the association is reflexive
            old_entity = biased_choice(entities)
            refs = [old_entity, old_entity]
        else:
            # Normal case: create a new entity
            new_entity = f"ENTITY_NAME_PLACEHOLDER {counter}_"
            new_ent_attrs = [f"{ent_attr_name} {counter} {j}" for (j, ent_attr_name) in enumerate(ent_attr_names[i], 1)]
            clauses.append(f"{new_entity}: {', '.join(new_ent_attrs)}")
            try: # Prefer all distinct entities
                old_refs = biased_sample(entities, arities[i] - 1)
            except ValueError: # If there are not enough entities, allow reflexive associations
                old_refs = [biased_choice(entities) for _ in range(arities[i] - 1)]
            refs = [new_entity] + old_refs
            entities.append(new_entity)
            counter += 1
        
        new_association = f"{association_bases[arities[i]]} {counter}_"
        ref_pool.append(refs)
        cards = calculate_cards(card_schemes[i], refs)
        clauses.append(", ".join([new_association] + [f"{card} {ref}" for (card, ref) in zip(cards, refs)]))
        weak_entities.update(extract_weak_entities(cards, refs))
        new_assoc_attrs = [f"{attr_base} {counter} {j}" for j in range(1, settings["assoc_attrs"][i] + 1)]
        if new_assoc_attrs:
            clauses[-1] += f": {', '.join(new_assoc_attrs)}"
        associations.append(new_association)
        counter += 1
    for i in range(settings["n"] - settings["doubles"], settings["n"]):
        refs = biased_choice(ref_pool)
        arity = len(set(refs))
        new_association = f"{association_bases[arity]} {counter}_"
        cards = calculate_cards(card_schemes[i], refs)
        clauses.append(", ".join([new_association] + [f"{card} {ref}" for (card, ref) in zip(cards, refs)]))
        weak_entities.update(extract_weak_entities(cards, refs))
        new_assoc_attrs = [f"{attr_base} {counter} {j}" for j in range(1, settings["assoc_attrs"][i] + 1)]
        if new_assoc_attrs:
            clauses[-1] += f": {', '.join(new_assoc_attrs)}"
        associations.append(new_association)
        counter += 1
    
    text = "\n".join(clauses)

    j = len("ENTITY_NAME_PLACEHOLDER")
    for e in weak_entities:
        text = re.sub(fr"\b{e}\b", lambda m: f"{weak_entity_base}{m.group(0)[j:]}", text)
    text = text.replace("ENTITY_NAME_PLACEHOLDER", entity_base)
    
    return text



from pathlib import Path
from collections import defaultdict
import json
import os

from mocodo.argument_parser import init_localization, Transformations
from mocodo.tools import load_mini_yaml

# Build the index of templates

folder = Path(f"mocodo/resources/relation_templates")
aliases = defaultdict(list)
metadata = defaultdict(dict)
for path in sorted(folder.glob("*.yaml")):
    if "-" in path.name:
        continue
    data = load_mini_yaml.run(path)
    if "help_en" in data:
        metadata[path.stem] = {
            "category": "cv",
            "help_en": data["help_en"],
            "help_fr": data["help_fr"],
            "help_zh": data["help_zh"],
            "aliases": [],
        }
        if "fr_examples" in data:
            fr_examples = {}
            for d in data["fr_examples"]:
                fr_examples[d["example"]] = d["explanation"]
            metadata[path.stem]["fr_examples"] = fr_examples
    elif "parent" in data:
        aliases[data["parent"]].append(path.stem)
    for (parent, children) in aliases.items():
        metadata[parent]["aliases"] = children
result = json.dumps(metadata, indent=2, ensure_ascii=False)
result = result.replace("[\n      ", "[")
result = result.replace("\n    ]", "]")
result = result.replace(",\n      ", ", ")
Path(folder, "_index.json").write_text(result + "\n", encoding="utf8")
print(f"File written: {folder}/_index.json")

# Build the graph of templates

edges = []
for path in sorted(folder.glob("*.yaml")):
    data = load_mini_yaml.run(path)
    data.pop("help_fr", None)
    data.pop("help_en", None)
    data.pop("fr_examples", None)
    color = min(9, len(data))
    font_color = "white" if color > 5 else "black"
    if "parent" in data:
        edges.append(f'"{data["parent"]}" -> "{path.stem}"')
    edges.append(f'"{path.stem}" [fillcolor={color} fontcolor={font_color}]')
edges = "\n  ".join(edges)
result = f'digraph G {{\n  rankdir=LR\n  edge [dir="back"]\n  node [shape=box style="rounded,filled" colorscheme=reds9 fontname="Arial" penwidth=0]\n  {edges}\n}}\n'
path = Path(folder, "_graph.gv")
path.write_text(result, encoding="utf8")
os.system(f"dot -Tsvg {folder}/_graph.gv > {folder}/_graph.svg")
path.unlink()
print(f"File written: {folder}/_graph.svg")

# Build the cheat sheet

init_localization("fr")
metadata = Transformations("fr").metadata
tables = {}
for category in ["rw", "cv"]:
    rows = []
    rows.append(("Sous-option", "Description", "Exemples", "Explications"))
    rows.append((":--", ":--", ":--", ":--"))
    for (option, data) in sorted(metadata.items()):
        if data["category"] != category:
            continue
        title = f" title=\"Alias : {', '.join(data['aliases'])}.\"" if data.get("aliases") else ""
        option = f'<span{title} style="font-family:monospace; font-weight:600">{option}</span>'
        row = [option, data['help']]
        if not data.get(f"fr_examples"):
            rows.append(row + ["", ""])
            continue
        for (example, description) in data[f"fr_examples"].items():
            example = example.replace("\n", "\\n")
            example = f"`` {example} ``"
            row.extend([example, description])
            rows.append(row)
            row = ["", ""]  # empty cells for the first two columns
    tables[category] = "\n".join("| " + " | ".join(row) + " |" for row in rows)
text = f"""
### Opérations de conversion

{tables['cv']}

### Opérations de réécriture

{tables['rw']}
""".strip()
output_path = Path("doc/fr_cheat_sheet.md")
output_path.write_text(text, encoding="utf8")
print(f"File written: {output_path}")



from ..parse_mcd import Visitor
from ..tools.parser_tools import parse_source, reconstruct_source, first_child
from ..tools.various import invert_dict


def fix_card(
    card,
    fixes=invert_dict(
        {
            "01": ["O1", "o1", "10", "1O", "1o", "Ol", "ol", "l0", "lO", "lo"],
            "0N": ["ON", "oN", "NO", "No", "N0"],
            "0n": ["On", "on", "no", "nO", "n0"],
            "1N": ["N1", "Nl"],
            "1n": ["n1", "nl"],
        }
    ),
):
    return fixes.get(card, card)

class DfInference(Visitor):

    def __init__(self, df_label):
        self.df_label = df_label

    def assoc_clause(self, tree):
        assoc_name = first_child(tree, "assoc_name_def").children[0]
        if assoc_name == self.df_label:
            return
        cards = []
        legs = list(tree.find_data("assoc_leg"))
        for leg in legs:
            card_prefix = first_child(leg, "card_prefix")
            if card_prefix == "/":
                return
            card = first_child(leg, "card")
            cards.append(card)
        if "11" in cards:
            assoc_name.value = self.df_label

def infer_dfs(source, df_label):
    tree = parse_source(source)
    visitor = DfInference(df_label)
    visitor.visit(tree)
    return reconstruct_source(tree)

class RoleInference(Visitor):

    def assoc_clause(self, tree):
        assoc_name = first_child(tree, "assoc_name_def").children[0]
        legs = list(tree.find_data("assoc_leg"))
        cards = [first_child(leg, "card") for leg in legs]
        roles = [first_child(leg, "leg_note") for leg in legs]
        if "01" in cards or "11" in cards:
            for (card, role) in zip(cards, roles):
                if role:
                    continue
                if card[1] != "1": # *1 vs *N
                    card.value += f" [{assoc_name}]"
                elif card == "01" and "11" in cards: # 01 vs 11
                    card.value += f" [{assoc_name}]"
                elif card == "11" and cards.count("11") > 1: # 11 vs 11
                    card.value += f" [{assoc_name}]"

def infer_roles(source):
    tree = parse_source(source)
    visitor = RoleInference()
    visitor.visit(tree)
    return reconstruct_source(tree)



import contextlib
from functools import lru_cache
from itertools import count, product
from math import hypot
from random import choice, random, shuffle

from .cross import cross
from ..mocodo_error import MocodoError
from ..argument_parser import non_negative_integer, positive_integer
from ..mcd import Mcd

def arrange(source, subargs, has_expired):

    is_organic = False
    if "wide" in subargs:
        col_count = 8
        with contextlib.suppress(Exception):
            col_count = int(subargs["wide"]) or 1
        mcd = Mcd(source)
        (q, r) = divmod(mcd.get_non_phantom_count(), col_count)
        row_count = q + bool(r)
        source = mcd.get_refitted_clauses(col_count, row_count)
    elif "balanced" in subargs:
        nth_fit = 0
        with contextlib.suppress(Exception):
            nth_fit = int(subargs["balanced"])
        mcd = Mcd(source)
        source = mcd.get_refitted_clauses(nth_fit)
    elif "current" in subargs:
        pass
    else:
        is_organic = True
    
    mcd = Mcd(source)
    layout_data = mcd.get_layout_data()
    successors = layout_data["successors"]
    col_count = layout_data["col_count"]
    row_count = layout_data["row_count"]
    multiplicity = layout_data["multiplicity"]

    min_objective = non_negative_integer(subargs.get("min_objective", 0))
    max_objective = non_negative_integer(subargs.get("max_objective", 15))
    call_limit = positive_integer(subargs.get("call_limit", 10000))
    verbose = bool(subargs.get("verbose") is not None) # -t arrange:verbose => {"verbose": ""} => True

    has_expired = has_expired or (lambda: False)

    @lru_cache
    def bounded_neighborhood(x1, y1):
        result = set()
        for x2 in range(
            max(0, x1 - radius),
            min(col_count, x1 + radius + 1),
        ):
            for y2 in range(
                max(0, y1 - radius + abs(x1 - x2)),
                min(row_count, y1 + radius - abs(x1 - x2) + 1),
            ):
                if x1 != x2 or y1 != y2:
                    result.add((x2, y2))
        return result

    @lru_cache
    def organic_neighborhood(x1, y1):
        result = set()
        for x2 in range(x1 - radius, x1 + radius + 1):
            for y2 in range(y1 - radius + abs(x1 - x2), y1 + radius - abs(x1 - x2) + 1):
                if x1 != x2 or y1 != y2:
                    result.add((x2, y2))
        return result

    @lru_cache
    def bounded_hull(coords):
        result = set()
        for (x, y) in coords:
            if x - 1 >= 0:
                result.add((x - 1, y))
            if x + 1 < col_count:
                result.add((x + 1, y))
            if y - 1 >= 0:
                result.add((x, y - 1))
            if y + 1 < row_count:
                result.add((x, y + 1))
        return result.difference(coords)

    @lru_cache
    def organic_hull(coords):
        result = set()
        for (x, y) in coords:
            result.add((x - 1, y))
            result.add((x + 1, y))
            result.add((x, y - 1))
            result.add((x, y + 1))
        return result.difference(coords)

    def recurs(box_coords, next_boxes, placed_segments, cumulated_distances):
        if cumulated_distances > objective:
            # print("cut")
            return None
        if len(next_boxes) == 0:
            return {
                "coords": box_coords,
                "crossings": 0,
                "distances": cumulated_distances,
            }
        outside_hull_count = len(next_boxes) - len(hull(frozenset(box_coords.values())))
        if outside_hull_count * outside_hull_minimal_distance + cumulated_distances > objective:
            # print("Lower bound cut")
            return None
        if has_expired():
            raise MocodoError(10, _('Layout calculation time exceeded.'))  # fmt: skip
        if next(iteration) > call_limit:
            # print("call limit exceeded")
            return None
        box_to_place = next_boxes[0]
        placed_successors = {
            box_coords[box]: box for box in successors[box_to_place] if box in box_coords
        }
        if placed_successors:
            placed_successor_coords = iter(placed_successors)
            (x1, y1) = next(placed_successor_coords)
            possible_coords = neighborhood(x1, y1).copy()
            # print(placed_successors[0], possible_coords)
            for (x1, y1) in placed_successor_coords:
                possible_coords.intersection_update(neighborhood(x1, y1))
                if not possible_coords:
                    # print("neighborhood intersection is empty")
                    return None
        else:
            # print("the box to place has no successors: all empty coords are possible")
            possible_coords = set(product(range(col_count), range(row_count)))
        possible_coords.difference_update(box_coords.values())
        if not possible_coords:
            # print("neighborhood intersection is not free")
            return None
        non_crossing_possible_coords = []
        for (x1, y1) in possible_coords:
            for ((x2, y2), (x3, y3, x4, y4)) in product(placed_successors, placed_segments):
                if cross(x1, y1, x2, y2, x3, y3, x4, y4):
                    break
            else:
                non_crossing_possible_coords.append((x1, y1))
        if not non_crossing_possible_coords:
            # print("all possible coords result in a crossing with existing segment")
            return None
        weighted_possible_coords = []
        for (x1, y1) in non_crossing_possible_coords:
            cumulated_distance = 0
            for ((x2, y2), placed_box) in placed_successors.items():
                cumulated_distance += (
                    distances[abs(x1 - x2)][abs(y1 - y2)] * multiplicity[(box_to_place, placed_box)]
                )
            weighted_possible_coords.append((cumulated_distance, random(), x1, y1))
        weighted_possible_coords.sort()
        for (cumulated_distance, __, x1, y1) in weighted_possible_coords:  # For the double underscore, see __main__.py
            box_coords[box_to_place] = (x1, y1)
            new_segments = [(x1, y1, x2, y2) for (x2, y2) in placed_successors]
            new_next_boxes = list(
                successors[box_to_place].difference(box_coords).difference(next_boxes)
            )
            if len(next_boxes) == 1 and len(new_next_boxes) == 0 and len(box_coords) != box_count:
                # print("the placed boxes have no more non placed successors")
                new_next_boxes = list(set(range(box_count)).difference(box_coords))
                if new_next_boxes:
                    new_next_boxes = [choice(new_next_boxes)]
            shuffle(new_next_boxes)
            result = recurs(
                box_coords,
                next_boxes[1:] + new_next_boxes,
                placed_segments + new_segments,
                cumulated_distances + cumulated_distance,
            )
            if result:
                return result
            del box_coords[box_to_place]

    box_count = col_count * row_count
    neighborhood = organic_neighborhood if is_organic else bounded_neighborhood
    hull = organic_hull if is_organic else bounded_hull
    radius = 3
    distances = [[hypot(i, j) - 1 for j in range(radius + 1)] for i in range(radius + 1)]
    outside_hull_minimal_distance = distances[1][2]
    if all(not successor for successor in successors):
        # print("no link: return a random layout")
        layout = list(range(box_count))
        shuffle(layout)
        result = {
            "layout": layout,
            "crossings": 0,
            "distances": 0,
        }
        mcd.set_layout(**result)
        return mcd.get_clauses()
    for objective in range(min_objective, max_objective + 1):
        if verbose:
            print("Objective %s." % objective)
        boxes = list(range(box_count))
        shuffle(boxes)
        for first_box in boxes:
            iteration = count()
            if successors[first_box]:
                if verbose:
                    print("  Starting from box %s." % first_box)
                result = recurs({first_box: (0, 0)}, list(successors[first_box]), [], 0)
                if result:
                    coords = result["coords"]
                    if is_organic:
                        min_x = min(x for (x, y) in coords.values())
                        max_x = max(x for (x, y) in coords.values())
                        min_y = min(y for (x, y) in coords.values())
                        max_y = max(y for (x, y) in coords.values())
                        for (box_index, (x, y)) in coords.items():
                            coords[box_index] = (x - min_x, y - min_y)
                        result["row_count"] = row_count = max_y - min_y + 1
                        result["col_count"] = col_count = max_x - min_x + 1
                    result["layout"] = [None] * row_count * col_count
                    for (box_index, (x, y)) in coords.items():
                        result["layout"][x + y * col_count] = box_index
                    mcd.set_layout(**result)
                    if mcd.row_count > mcd.col_count:
                        # If the layout is taller than wide, transpose it.
                        return mcd.get_diagonally_flipped_clauses()
                    else:
                        return mcd.get_clauses()
                if is_organic:
                    break
        objective += 1
    if is_organic:
        raise MocodoError(41, _('Failed to calculate a non-constrained planar layout.'))  # fmt: skip
    else:
        raise MocodoError(9, _('Failed to calculate a planar layout satisfying the given constraint.'))  # fmt: skip


if __name__ == "__main__":
    # python -m mocodo.rewrite._arrange_bb

    from time import time
    import random
    from ..argument_parser import parsed_arguments
    from ..mcd import Mcd

    clauses = """
        SUSPENDISSE: diam
        SOLLICITUDIN, 0N SUSPENDISSE, 0N CONSECTETUER, 0N LOREM: lectus
        CONSECTETUER: elit, sed
        MAECENAS, 1N DIGNISSIM, 1N DIGNISSIM

        DF, 11 LOREM, 1N SUSPENDISSE
        LOREM: ipsum, dolor, sit
        TORTOR, 0N RISUS, 11 DIGNISSIM, 1N CONSECTETUER: nec
        DIGNISSIM: ligula, massa, varius

        DF, 11 RISUS, 0N RISUS
        AMET, 11> LOREM, 01 CONSECTETUER: adipiscing
        RISUS: ultricies, _cras, elementum
        SEMPER, 0N RISUS, 1N DIGNISSIM
    """
    params = parsed_arguments()
    mcd = Mcd(clauses.replace("  ", ""), **params)
    starting_time = time()
    random.seed(42)
    rearrangement = arrange(mcd, {"is_organic": True}, lambda: False)
    if rearrangement:
        print()
        mcd.set_layout(**rearrangement)
        print(mcd.get_clauses())
        print()
        print("Cumulated distances:", rearrangement["distances"])
        print("Duration:", time() - starting_time)
        print()



class Phantom:

    counter = 0

    @classmethod
    def reset_counter(cls):
        cls.counter = 0

    def __init__(self):
        self.source = ":"
        self.name_view = ""
        Phantom.counter += 1
        self.bid = f"PHANTOM_#{Phantom.counter}"
        self.attributes = []
        self.legs = [] # iterating over box's legs does nothing if it is not an association
        self.kind = "phantom"
        self.page = 0

    def calculate_size(self, *ignored):
        self.w = 0
        self.h = 0

    def register_center(self, geo):
        pass

    def register_boxes(self, boxes):
        self.boxes = boxes



import importlib

def load_ipython_extension(ipython):
    # This function is called when the extension is loaded in a notebook
    # with %load_ext mocodo or %reload_ext mocodo.

    mocodo = importlib.import_module("mocodo.magic_command").mocodo
    ipython.register_magic_function(mocodo, 'line_cell', 'mocodo')
    
    pkg_resources = importlib.import_module("pkg_resources")

    try:
        MOCODO_VERSION = pkg_resources.get_distribution("mocodo").version
    except pkg_resources.DistributionNotFound:
        MOCODO_VERSION = "(unknown version)"  # For tests during CI
    print(f"Mocodo {MOCODO_VERSION} loaded.")



import re
from ..parse_mcd import Visitor
from ..tools.parser_tools import first_child, parse_source
from ..tools.string_tools import surrounds
from ..tools.various import first_missing_positive
    
class CreateCifs(Visitor):

    def __init__(self):
        self.candidates = set()
        self.existing = set()
        self.invisible_numbers = []

    def entity_clause(self, tree):
        ent_name = str(first_child(tree, "box_name"))
        m = re.match(r"(?i)invisible(\d+)$", ent_name)
        if m:
            self.invisible_numbers.append(int(m.group(1)))

    def assoc_clause(self, tree):
        legs = [node for node in tree.find_data("assoc_leg")]
        assoc_name = str(first_child(tree, "assoc_name_def").children[0])
        target_names = []
        entity_names = set()
        for leg in legs:
            entity_name = first_child(leg, "entity_name_ref").children[0].value
            entity_names.add(entity_name)
            if first_child(leg, "card_prefix") == "/":
                target_names.append(entity_name)
        for target_name in target_names:
            self.candidates.add((assoc_name, target_name, *sorted(entity_names - {target_name})))
    
    def constraint_clause(self, tree):
        if first_child(tree, "constraint_name").upper() != "CIF":
            return
        assoc_name = None
        target_name = None
        source_names = []
        for node in tree.find_data("constraint_target"):
            leg = first_child(node, "constraint_leg")
            box_name = first_child(node, "box_name_ref").children[0].value
            if surrounds(leg, ".."):
                if assoc_name is not None:
                    return
                assoc_name = box_name
            elif surrounds(leg, "->"):
                if target_name is not None:
                    return
                target_name = box_name
            elif leg == "" or surrounds(leg, "--"):
                source_names.append(box_name)
            else:
                return
        self.existing.add((assoc_name, target_name, *sorted(source_names)))
    
    def get_new_cifs(self, hidden_links_from_sources=False):
        self.new_cifs = self.candidates - self.existing
        if not self.new_cifs:
            return ""
        sep = ", " if hidden_links_from_sources else ", --"
        new_invisible_clauses = ["\n"]
        new_cif_clauses = [""]
        for (assoc_name, *entity_names) in sorted(self.new_cifs):
            n = first_missing_positive(self.invisible_numbers)
            self.invisible_numbers.append(n)
            new_invisible_clauses.append(f"-INVISIBLE_{n}, XX {entity_names[0]}, XX {entity_names[0]}")
            new_cif_clauses.append(f"(CIF) ..{assoc_name}, ->{sep.join(entity_names)}: INVISIBLE_{n}, INVISIBLE_{n}")
        if new_cif_clauses == [""]:
            return ""
        return "\n".join(new_invisible_clauses + new_cif_clauses)


def create_cifs(source, subsubarg):
    visitor = CreateCifs()
    tree = parse_source(source)
    visitor.visit(tree)
    return source + visitor.get_new_cifs(subsubarg == "light")



__import__("sys").path[0:0] = ["."]

from ._crow import Crow
from ..tools.parser_tools import parse_source
from ..rewrite import (
    _drain as drain,
    _explode as explode,
    _split as split,
)
from ..tools.string_tools import rstrip_digit_or_underline
from ..tools.graphviz_tools import NODE_OPTIONS_TEMPLATE, table_as_label
from ..version_number import version

SUFFIX = "_crows_foot_erd.gv"

GV_CARD = {
    "01": "teeodot",
    "0N": "crowodot",
    "11": "teetee",
    "1N": "crowtee",
}

class CrowGv(Crow):
  
    def get_text(self, common):
        style = common.load_style()
        ent_table_style = {
            "stroke_color": style["entity_stroke_color"],
            "cell_bg_color": style["entity_color"],
            "header_bg_color": style["entity_cartouche_color"],
            "header_font_color": style["entity_cartouche_text_color"],
            "header_font_size": style["entity_cartouche_font"]["size"],
            "cell_font_color": style["entity_attribute_text_color"],
            "cell_font_size": style["entity_attribute_font"]["size"],
        }
        assoc_table_style = {
            "stroke_color": style["association_stroke_color"],
            "cell_bg_color": style["association_color"],
            "header_bg_color": style["association_cartouche_color"],
            "header_font_color": style["association_cartouche_text_color"],
            "header_font_size": style["association_cartouche_font"]["size"],
            "cell_font_color": style["association_attribute_text_color"],
            "cell_font_size": style["association_attribute_font"]["size"],
        }
        acc = []
        acc.append(f'// Generated by Mocodo {version}\n')
        acc.append(f'digraph{{')
        acc.append(f'  layout=dot')
        acc.append(f'  bgcolor="{style["background_color"]}"')
        acc.append(f'  nodesep=0.5') # increase spacing between multiedges

        acc.append(f'\n  // Nodes')
        for (i, table_style) in enumerate([assoc_table_style, ent_table_style]):
            node_options = NODE_OPTIONS_TEMPLATE.format(**ent_table_style)
            acc.append(f'  node [{node_options}]')
            for (ent_index, (ent_name, has_id, attrs)) in self.tables.items():
                if i != has_id:
                    continue
                attrs = [(("PK" if is_id else " "), a, (t or " ")) for (t, a, is_id) in attrs]
                row_format = "|c|lR|"
                if self.has_no_datatype:
                    attrs = [attr[:-1] for attr in attrs]
                    row_format = row_format.replace("R|", "|")
                label = table_as_label(ent_name, attrs, row_format, table_style)
                acc.append(f'  {ent_index} [label=<{label}>]')
        
        acc.append(f'\n  // Edges')
        acc.append(f'  edge [')
        acc.append(f'    penwidth={style["leg_stroke_depth"]}')
        acc.append(f'    color="{style["leg_stroke_color"]}"')
        acc.append(f'    fontcolor="{style["card_text_color"]}"')
        acc.append(f'    fontname="{style["card_font"]["family"]}"')
        acc.append(f'    fontsize={style["card_font"]["size"]}')
        acc.append(f'    dir=both') # bidirectional arrows, otherwise the tail doesn't appear
        acc.append(f'  ]')
        for (ent_1, card_1, kind, card_2, ent_2, assoc_name) in self.links:
            ent_index_1 = self.name_to_index(ent_1)
            ent_index_2 = self.name_to_index(ent_2)
            tail = GV_CARD.get(card_1, "crowodot")
            head = GV_CARD.get(card_2, "crowodot")
            kind = " style=dotted" if kind == ".." else ""
            label = ""
            if self.tables[ent_index_1][1] and self.tables[ent_index_2][1]:
                label = f' label="{rstrip_digit_or_underline(assoc_name)}"'
            acc.append(f'  {ent_index_1} -> {ent_index_2} [arrowhead="{head}" arrowtail="{tail}"{label}{kind}]')
        acc.append('}')

        return "\n".join(acc)

def run(source, subargs, common=None):
    source = drain.run(source)
    source = split.run(source)
    source = explode.run(source, {"arity": "2.5", "weak": True}, common.params)
    tree = parse_source(source)
    extractor = CrowGv()
    extractor.visit(tree)
    result = extractor.get_text(common)
    return result



import operator
from math import hypot, sqrt

from .mocodo_error import MocodoError
from .tools.string_tools import surrounds, raw_to_bid


class Leg:
    def __init__(self, association, leg, **params):
        params["card_format"] = params.get("card_format", "{min_card},{max_card}")

        self.card = leg.get("card", "XX")
        if self.card == "XX" or leg.get("card_hidden") == "-":
            self.card_view = "     "
        elif "X" in self.card:
            self.card_view = self.card.replace("X", "")
        else:
            self.card_view = params["card_format"].format(min_card=self.card[0], max_card=self.card[1])
        
        self.has_underlined_card = False
        if leg.get("card_prefix") == "_":
            if self.card != "11": # silently ignore the prefix
                self.kind = "leg"
                del leg["card_prefix"]
            else:
                self.kind = "strengthening"
                self.card_view = params.get("strengthen_card", "_1,1_")
                if surrounds(self.card_view, "_"):
                    self.has_underlined_card = True
                    self.card_view = self.card_view[1:-1]
        elif association.kind == "cluster":
            self.kind = "cluster_peg" if leg.get("card_prefix") == "/" else "cluster_leg"
        else:
            self.kind = "leg"
        
        self.arrow = leg.get("leg_arrow", "")
        self.peg = ("o" if self.kind == "cluster_peg" and not self.arrow else "")
        self.note = leg.get("leg_note")
        self.association = association
        self.entity_raw_name = leg["entity"]
        self.entity_bid = raw_to_bid(self.entity_raw_name)
        self.twist = False
        self.lid = None
        self.unicities = ""
        self.is_in_elected_group = False

    def register_entity(self, entity):
        self.entity = entity
    
    def register_mcd_has_cif(self, mcd_has_cif):
        self.mcd_has_cif = mcd_has_cif
        if self.mcd_has_cif:
            self.peg = ""

    def append_candidate_group(self, candidate_number: str):
        if candidate_number == "0":
            self.is_in_elected_group = True
        else:
            # concatenate the candidate number to the beginning of the string to ensure sorting
            self.unicities = candidate_number + self.unicities

    def calculate_size(self, style, get_font_metrics):
        font = get_font_metrics(style["card_font"])
        self.h = font.get_pixel_height()
        self.w = font.get_pixel_width(self.card_view)

    def set_spin_strategy(self, spin):
        self.spin = spin
        self.description = self._curved_description if spin else self._straight_description
        if self.lid is None:
            spin_str = str(round(spin, 2)).replace(".", "_")  # avoid dot for the web version
            self.lid = f"{self.association.bid},{self.entity_bid},{spin_str}"

    def _straight_description(self, style, geo):
        result = []
        ex = self.entity.cx
        ey = self.entity.cy
        ew = self.entity.w // 2
        eh = self.entity.h // 2
        ax = self.association.cx
        ay = self.association.cy
        aw = self.association.w // 2
        ah = self.association.h // 2
        card_margin = style["card_margin"]
        cw = self.w + 2 * card_margin
        ch = self.h + 2 * card_margin
        leg = straight_leg_factory(ex, ey, ew, eh, ax, ay, aw, ah, cw, ch, card_margin)
        if self.peg:
            (x, y, a, b) = leg.arrow_pos(self.arrow, 1)
            c = hypot(a, b)
            (cos, sin) = (a / c, b / c)
            result.append(
                (
                    "circle",
                    {
                        "cx": x + cos * style["arrow_width"] / 2,
                        "cy": y - sin * style["arrow_width"] / 2,
                        "r": style["arrow_width"] / 2,
                        "stroke_color": style["entity_stroke_color"],
                        "stroke_depth": style["leg_stroke_depth"],
                        "color": style["entity_color"]+ "55", # alpha channel
                    },
                ),
            )
        result.append(
            (
                "line",
                {
                    "x0": ex,
                    "y0": ey,
                    "x1": ax,
                    "y1": ay,
                    "stroke_color": style["leg_stroke_color"],
                    "stroke_depth": style["leg_stroke_depth"],
                },
            )
        )
        (x, y) = leg.card_pos(self.twist, geo["shift"][self.lid])
        tx = x + card_margin
        ty = y - card_margin - style["card_baseline"]
        self.saved_card_description = []
        if self.note:
            self.saved_card_description.append(
                (
                    "text_with_note",
                    {
                        "x": tx,
                        "y": ty,
                        "text_color": style["card_text_color"],
                        "family": style["card_font"]["family"],
                        "size": style["card_font"]["size"],
                        "text": self.card_view,
                        "note": self.note.lstrip("+").lstrip("-")
                    },
                )
            )
        else:
            self.saved_card_description.append(
                (
                    "text",
                    {
                        "x": tx,
                        "y": ty,
                        "text_color": style["card_text_color"],
                        "family": style["card_font"]["family"],
                        "size": style["card_font"]["size"],
                        "text": self.card_view,
                    },
                )
            )
        if self.has_underlined_card:
            self.saved_card_description.append(
                (
                    "line",
                    {
                        "x0": tx,
                        "y0": ty - style["card_underline_skip_height"],
                        "x1": tx + self.w,
                        "y1": ty - style["card_underline_skip_height"],
                        "stroke_color": style["card_text_color"],
                        "stroke_depth": style["card_underline_depth"],
                    },
                )
            )
        if self.arrow:
            (x, y, a, b) = leg.arrow_pos(self.arrow, geo["ratio"][self.lid])
            c = hypot(a, b)
            (cos, sin) = (a / c, b / c)
            result.append(
                (
                    "arrow",
                    {
                        "x0": x,
                        "y0": y,
                        "x1": x + style["arrow_width"] * cos - style["arrow_half_height"] * sin,
                        "y1": y - style["arrow_half_height"] * cos - style["arrow_width"] * sin,
                        "x2": x + style["arrow_axis"] * cos,
                        "y2": y - style["arrow_axis"] * sin,
                        "x3": x + style["arrow_width"] * cos + style["arrow_half_height"] * sin,
                        "y3": y + style["arrow_half_height"] * cos - style["arrow_width"] * sin,
                        "stroke_color": style["leg_stroke_color"],
                    },
                ),
            )
        return result

    def _curved_description(self, style, geo):
        result = []
        ex = self.entity.cx
        ey = self.entity.cy
        ew = self.entity.w // 2
        eh = self.entity.h // 2
        ax = self.association.cx
        ay = self.association.cy
        aw = self.association.w // 2
        ah = self.association.h // 2
        card_margin = style["card_margin"]
        cw = self.w + 2 * card_margin
        ch = self.h + 2 * card_margin
        spin = self.spin
        leg = curved_leg_factory(ex, ey, ew, eh, ax, ay, aw, ah, cw, ch, card_margin, spin)
        (x0, y0, x1, y1, x2, y2, x3, y3) = leg.points
        if self.peg:
            (x, y, a, b) = leg.arrow_pos(self.arrow, 1)
            c = hypot(a, b)
            (cos, sin) = (a / c, b / c)
            result.append(
                (
                    "circle",
                    {
                        "cx": x + cos * style["arrow_width"] / 2,
                        "cy": y - sin * style["arrow_width"] / 2,
                        "r": style["arrow_width"] / 2,
                        "stroke_color": style["entity_stroke_color"],
                        "stroke_depth": style["leg_stroke_depth"],
                        "color": style["entity_color"]+ "55", # alpha channel
                    },
                ),
            )
        result.append(
            (
                "curve",
                {
                    "x0": x0,
                    "y0": y0,
                    "x1": x1,
                    "y1": y1,
                    "x2": x2,
                    "y2": y2,
                    "x3": x3,
                    "y3": y3,
                    "stroke_color": style["leg_stroke_color"],
                    "stroke_depth": style["leg_stroke_depth"],
                },
            )
        )
        (x, y) = leg.card_pos(geo["shift"][self.lid])
        tx = x + card_margin
        ty = y - card_margin - style["card_baseline"]
        self.saved_card_description = []
        if self.note:
            self.saved_card_description.append(
                (
                    "text_with_note",
                    {
                        "x": tx,
                        "y": ty,
                        "text_color": style["card_text_color"],
                        "family": style["card_font"]["family"],
                        "size": style["card_font"]["size"],
                        "text": self.card_view,
                        "note": self.note.lstrip("+").lstrip("-"),
                    },
                )
            )
        else:
            self.saved_card_description.append(
                (
                    "text",
                    {
                        "x": tx,
                        "y": ty,
                        "text_color": style["card_text_color"],
                        "family": style["card_font"]["family"],
                        "size": style["card_font"]["size"],
                        "text": self.card_view,
                    },
                )
            )
        if self.has_underlined_card:
            self.saved_card_description.append(
                (
                    "line",
                    {
                        "x0": tx,
                        "y0": ty - style["card_underline_skip_height"],
                        "x1": tx + self.w,
                        "y1": ty - style["card_underline_skip_height"],
                        "stroke_color": style["card_text_color"],
                        "stroke_depth": style["card_underline_depth"],
                    },
                )
            )
        if self.arrow:
            (x, y, a, b) = leg.arrow_pos(self.arrow, geo["ratio"][self.lid])
            c = hypot(a, b)
            (cos, sin) = (a / c, b / c)
            result.append(
                (
                    "arrow",
                    {
                        "x0": x,
                        "y0": y,
                        "x1": x + style["arrow_width"] * cos - style["arrow_half_height"] * sin,
                        "y1": y - style["arrow_half_height"] * cos - style["arrow_width"] * sin,
                        "x2": x + style["arrow_axis"] * cos,
                        "y2": y - style["arrow_axis"] * sin,
                        "x3": x + style["arrow_width"] * cos + style["arrow_half_height"] * sin,
                        "y3": y + style["arrow_half_height"] * cos - style["arrow_width"] * sin,
                        "stroke_color": style["leg_stroke_color"],
                    },
                ),
            )
        return result

class InheritanceLeg:
    def __init__(self, inheritance, leg, **params):
        self.kind = leg["kind"] # "-" or "="
        self.arrow = leg["arrow"]
        self.inheritance = inheritance
        self.entity_bid = raw_to_bid(leg["entity"])
        self.entity_raw_name = leg["entity"]
        self.lid = f"{self.inheritance.bid} / {self.entity_bid}"

    def register_entity(self, entity):
        self.entity = entity
    
    def calculate_size(self, style, get_font_metrics):
        font = get_font_metrics(style["card_font"])
        self.h = font.get_pixel_height()
        self.w = font.get_pixel_width("--")

    def description(self, style, geo):
        result = []
        ex = self.entity.cx
        ey = self.entity.cy
        ew = self.entity.w // 2
        eh = self.entity.h // 2
        ax = self.inheritance.cx
        ay = self.inheritance.cy
        aw = self.inheritance.w // 2
        ah = self.inheritance.h // 2
        card_margin = style["card_margin"]
        cw = self.w + 2 * card_margin
        ch = self.h + 2 * card_margin
        leg = straight_leg_factory(ex, ey, ew, eh, ax, ay, aw, ah, cw, ch, card_margin)
        if self.kind == "=":
            for d in (style["leg_stroke_depth"], -style["leg_stroke_depth"]):
                (x0, y0, x1, y1) = orthogonal_translation(ex, ey, ax, ay, d)
                result.append(
                    (
                        "line",
                        {
                            "x0": x0,
                            "y0": y0,
                            "x1": x1,
                            "y1": y1,
                            "stroke_color": style["leg_stroke_color"],
                            "stroke_depth": style["leg_stroke_depth"],
                        },
                    )
                )
        else:
            result.append(
                (
                    "line",
                    {
                        "x0": ex,
                        "y0": ey,
                        "x1": ax,
                        "y1": ay,
                        "stroke_color": style["leg_stroke_color"],
                        "stroke_depth": style["leg_stroke_depth"],
                    },
                )
            )
        if self.arrow:
            (x, y, a, b) = leg.arrow_pos(self.arrow, geo["ratio"][self.lid])
            c = hypot(a, b)
            (cos, sin) = (a / c, b / c)
            result.append(
                (
                    "arrow",
                    {
                        "x0": x,
                        "y0": y,
                        "x1": x + style["arrow_width"] * cos - style["arrow_half_height"] * sin,
                        "y1": y - style["arrow_half_height"] * cos - style["arrow_width"] * sin,
                        "x2": x + style["arrow_axis"] * cos,
                        "y2": y - style["arrow_axis"] * sin,
                        "x3": x + style["arrow_width"] * cos + style["arrow_half_height"] * sin,
                        "y3": y + style["arrow_half_height"] * cos - style["arrow_width"] * sin,
                        "stroke_color": style["leg_stroke_color"],
                    },
                ),
            )
        return result


class ConstraintLeg:

    def __init__(self, constraint, kind, box_raw_name):
        self.constraint = constraint
        self.kind = kind
        self.bid = raw_to_bid(box_raw_name)
        self.entity_raw_name = box_raw_name
        self.lid = None
    
    def register_box(self, box):
        self.box = box
    
    def description(self, style, geo):
        if self.kind == "": # a phantom leg, useful to tweak the barycenter
            return []
        result = []
        bx = self.box.cx
        by = self.box.cy
        bw = self.box.w // 2
        bh = self.box.h // 2
        cx = self.constraint.cx
        cy = self.constraint.cy
        cw = self.constraint.w // 2
        ch = self.constraint.h // 2
        leg = straight_leg_factory(bx, by, bw, bh, cx, cy, cw, ch)
        kind = self.kind
        for direction in "<>":
            if direction in self.kind:
                (x, y, a, b) = leg.arrow_pos(direction, 1)
                c = hypot(a, b) or 1
                (cos, sin) = (a / c, b / c)
                result.append(
                    (
                        "arrow",
                        {
                            "x0": x,
                            "y0": y,
                            "x1": x + style["arrow_width"] * cos - style["arrow_half_height"] * sin,
                            "y1": y - style["arrow_half_height"] * cos - style["arrow_width"] * sin,
                            "x2": x + style["arrow_axis"] * cos,
                            "y2": y - style["arrow_axis"] * sin,
                            "x3": x + style["arrow_width"] * cos + style["arrow_half_height"] * sin,
                            "y3": y + style["arrow_half_height"] * cos - style["arrow_width"] * sin,
                            "stroke_color": style["constraint_stroke_color"],
                        },
                    ),
                )
                kind = kind.replace(direction, "")
        if kind[:1] == "-":
            result.append(
                (
                    "line",
                    {
                        "x0": bx,
                        "y0": by,
                        "x1": cx,
                        "y1": cy,
                        "stroke_color": style["constraint_stroke_color"],
                        "stroke_depth": style["constraint_stroke_depth"],
                    }
                )
            )
        elif kind[:1] == ".":
            result.append(
                (
                    "dot_line",
                    {
                        "x0": bx,
                        "y0": by,
                        "x1": cx,
                        "y1": cy,
                        "stroke_color": style["constraint_stroke_color"],
                        "stroke_depth": style["constraint_dot_stroke_depth"],
                        "dash_gap": style["constraint_dot_gap_width"],
                    }
                )
            )
        else:
            raise NotImplementedError
        return result


def line_intersection(ex, ey, w, h, ax, ay, cmp=lambda x, y: (x > y) - (x < y)):
    if ax == ex:
        return (ax, ey + cmp(ay, ey) * h)
    if ay == ey:
        return (ex + cmp(ax, ex) * w, ay)
    x = ex + cmp(ax, ex) * w
    y = ey + (ay - ey) * (x - ex) / (ax - ex)
    if abs(y - ey) > h:
        y = ey + cmp(ay, ey) * h
        x = ex + (ax - ex) * (y - ey) / (ay - ey)
    return (x, y)


def straight_leg_factory(ex, ey, ew, eh, ax, ay, aw, ah, cw=0, ch=0, card_margin=0):
    def card_pos(twist, shift):
        compare = operator.lt if twist else operator.le
        correction = 1 - abs(abs(ax - ex) - abs(ay - ey)) / hypot(ax - ex, ay - ey)
        correction = card_margin * 1.4142 * correction - shift
        (xg, yg) = line_intersection(ex, ey, ew, eh + ch, ax, ay)
        (xb, yb) = line_intersection(ex, ey, ew + cw, eh, ax, ay)
        if compare(xg, xb):
            if compare(xg, ex):
                if compare(yb, ey):
                    return (xb - correction, yb)
                else:
                    return (xb - correction, yb + ch)
            else:
                if compare(yb, ey):
                    return (xg, yg + ch - correction)
                else:
                    return (xg, yg + correction)
        else:
            if compare(xb, ex):
                if compare(yb, ey):
                    return (xg - cw, yg + ch - correction)
                else:
                    return (xg - cw, yg + correction)
            else:
                if compare(yb, ey):
                    return (xb - cw + correction, yb)
                else:
                    return (xb - cw + correction, yb + ch)

    def arrow_pos(direction, ratio):
        (x0, y0) = line_intersection(ex, ey, ew, eh, ax, ay)
        (x1, y1) = line_intersection(ax, ay, aw, ah, ex, ey)
        if direction == "<":
            (x0, y0, x1, y1) = (x1, y1, x0, y0)
        (x, y) = (ratio * x0 + (1 - ratio) * x1, ratio * y0 + (1 - ratio) * y1)
        return (x, y, x1 - x0, y0 - y1)

    straight_leg_factory.card_pos = card_pos
    straight_leg_factory.arrow_pos = arrow_pos
    return straight_leg_factory


def curved_leg_factory(ex, ey, ew, eh, ax, ay, aw, ah, cw, ch, card_margin, spin):
    def bisection(predicate):
        (a, b) = (0, 1)
        while abs(b - a) > 0.0001:
            m = (a + b) / 2
            if predicate(bezier(m)):
                a = m
            else:
                b = m
        return m

    def intersection(left, top, right, bottom):
        (x, y) = bezier(bisection(lambda p: left <= p[0] <= right and top <= p[1] <= bottom))
        return (int(round(x)), int(round(y)))  # avoid comparing floats

    def card_pos(shift):
        diagonal = hypot(ax - ex, ay - ey)
        correction = card_margin * 1.4142 * (1 - abs(abs(ax - ex) - abs(ay - ey)) / diagonal)
        (top, bot) = (ey - eh, ey + eh)
        (TOP, BOT) = (top - ch, bot + ch)
        (lef, rig) = (ex - ew, ex + ew)
        (LEF, RIG) = (lef - cw, rig + cw)
        (xr, yr) = intersection(LEF, TOP, RIG, BOT)
        (xg, yg) = intersection(lef, TOP, rig, BOT)
        (xb, yb) = intersection(LEF, top, RIG, bot)
        if spin > 0:
            if (yr == BOT and xr <= rig) or (xr == LEF and yr >= bot):
                return (
                    max(x for (x, y) in ((xr, yr), (xg, yg), (xb, yb)) if y >= bot)
                    - correction
                    + shift,
                    bot + ch,
                )
            if (xr == RIG and yr >= top) or yr == BOT:
                return (
                    rig,
                    min(y for (x, y) in ((xr, yr), (xg, yg), (xb, yb)) if x >= rig)
                    + correction
                    + shift,
                )
            if (yr == TOP and xr >= lef) or xr == RIG:
                return (
                    min(x for (x, y) in ((xr, yr), (xg, yg), (xb, yb)) if y <= top)
                    + correction
                    + shift
                    - cw,
                    TOP + ch,
                )
            return (
                LEF,
                max(y for (x, y) in ((xr, yr), (xg, yg), (xb, yb)) if x <= lef)
                - correction
                + shift
                + ch,
            )
        if (yr == BOT and xr >= lef) or (xr == RIG and yr >= bot):
            return (
                min(x for (x, y) in ((xr, yr), (xg, yg), (xb, yb)) if y >= bot)
                + correction
                + shift
                - cw,
                bot + ch,
            )
        if xr == RIG or (yr == TOP and xr >= rig):
            return (
                rig,
                max(y for (x, y) in ((xr, yr), (xg, yg), (xb, yb)) if x >= rig)
                - correction
                + shift
                + ch,
            )
        if yr == TOP or (xr == LEF and yr <= top):
            return (
                max(x for (x, y) in ((xr, yr), (xg, yg), (xb, yb)) if y <= top)
                - correction
                + shift,
                TOP + ch,
            )
        return (
            LEF,
            min(y for (x, y) in ((xr, yr), (xg, yg), (xb, yb)) if x <= lef) + correction + shift,
        )

    def arrow_pos(direction, ratio):
        t0 = bisection(lambda p: abs(p[0] - ax) > aw or abs(p[1] - ay) > ah)
        t3 = bisection(lambda p: abs(p[0] - ex) < ew and abs(p[1] - ey) < eh)
        if direction == "<":
            (t0, t3) = (t3, t0)
        tc = t0 + (t3 - t0) * ratio
        (xc, yc) = bezier(tc)
        (x, y) = derivate(tc)
        if direction == "<":
            (x, y) = (-x, -y)
        return (xc, yc, x, -y)

    diagonal = hypot(ax - ex, ay - ey)
    (x, y) = line_intersection(ex, ey, ew + cw / 2, eh + ch / 2, ax, ay)
    k = cw * abs((ay - ey) / diagonal) + ch * abs((ax - ex) / diagonal)
    (x, y) = (x - spin * k * (ay - ey) / diagonal, y + spin * k * (ax - ex) / diagonal)
    (hx, hy) = (2 * x - (ex + ax) / 2, 2 * y - (ey + ay) / 2)
    (x1, y1) = (ex + (hx - ex) * 2 / 3, ey + (hy - ey) * 2 / 3)
    (x2, y2) = (ax + (hx - ax) * 2 / 3, ay + (hy - ay) * 2 / 3)
    (kax, kay) = (ex - 2 * hx + ax, ey - 2 * hy + ay)
    (kbx, kby) = (2 * hx - 2 * ex, 2 * hy - 2 * ey)
    bezier = lambda t: (kax * t * t + kbx * t + ex, kay * t * t + kby * t + ey)
    derivate = lambda t: (2 * kax * t + kbx, 2 * kay * t + kby)

    curved_leg_factory.points = (ex, ey, x1, y1, x2, y2, ax, ay)
    curved_leg_factory.card_pos = card_pos
    curved_leg_factory.arrow_pos = arrow_pos
    return curved_leg_factory

def orthogonal_translation(x1, y1, x2, y2, d):
    if x1 == x2:
        return (x1+d, y1, x1+d, y2)
    slope = (y2 - y1) / (x2 - x1)
    dy = sqrt(d * d / (slope * slope + 1))
    dx = -slope * dy
    if d > 0:
        return (x1+dx, y1+dy, x2+dx, y2+dy)
    else:
        return (x1-dx, y1-dy, x2-dx, y2-dy)



from functools import lru_cache

crossed_strings = frozenset(["-++-", "-++0", "-+0-", "-0+-", "0++-", "+--+", "0--+", "+0-+", "+-0+", "+--0"])

@lru_cache
def cross(x1, y1, x2, y2, x3, y3, x4, y4):
    """ Tests whether the segments ((x1,y1), (x2,y2)) and ((x3,y3), (x4,y4)) intersect.
        Two segments sharing exactly one extremity are NOT considered as intersecting. """
    a = (x4-x3)*(y1-y3) - (y4-y3)*(x1-x3)
    b = (x4-x3)*(y2-y3) - (y4-y3)*(x2-x3)
    c = (x2-x1)*(y3-y1) - (y2-y1)*(x3-x1)
    d = (x2-x1)*(y4-y1) - (y2-y1)*(x4-x1)
    if a or b or c or d:
        return "".join("+" if x>0 else ("-" if x<0 else "0") for x in (a, b, c, d)) in crossed_strings
    else: # the segments are collinear
        if x1 == x2:  # both segments are vertical
            if y1 < y2:
                if y3 < y4:
                    return y2 > y3 and y1 < y4
                else:
                    return y2 > y4 and y1 < y3
            else:
                if y3 < y4:
                    return y1 > y3 and y2 < y4
                else:
                    return y1 > y4 and y2 < y3
        else:
            if x1 < x2:
                if x3 < x4:
                    return x2 > x3 and x1 < x4
                else:
                    return x2 > x4 and x1 < x3
            else:
                if x3 < x4:
                    return x1 > x3 and x2 < x4
                else:
                    return x1 > x4 and x2 < x3



from ..tools.string_tools import urlsafe_encoding


def run(source, subargs=None, common=None):
    return {
        "stem_suffix": "_url",
        "text": f"https://www.mocodo.net/?mcd={urlsafe_encoding(source)}",
        "extension": "url",
        "to_defer": True,
        "highlight": "url",
    }



from collections import namedtuple
from random import choice, random, randrange, sample

from .fitness import fitness
from ..argument_parser import rate, positive_integer
from ..mcd import Mcd

def arrange(source, subargs, has_expired):
    
    mcd = Mcd(source)
    layout_data = mcd.get_layout_data()
    links = layout_data["links"]
    successors = layout_data["successors"]
    col_count = layout_data["col_count"]
    row_count = layout_data["row_count"]
    multiplicity = layout_data["multiplicity"]

    verbose = subargs.get("verbose", False)
    population_size = positive_integer(subargs.get("population_size", 1000)) # number of individuals to evolve
    max_generations = positive_integer(subargs.get("max_generations", 300))
    plateau = positive_integer(subargs.get("plateau", 30)) # maximal number of consecutive generations without improvement
    crossover_rate = rate(subargs.get("crossover_rate", 0.9))
    mutation_rate = rate(subargs.get("mutation_rate", 0.06))
    sample_size = positive_integer(subargs.get("sample_size", 7))
    verbose = bool(subargs.get("verbose") is not None) # -r arrange:verbose => {"verbose": ""} => True

    def make_individual():
        """ Construct a chromosome. Select a random node for the first gene. The next ones are chosen
            sequentially. When a gene has another gene to the west, the corresponding node is preferably
            selected among the successors of the latter. NB: Applying the same technic for the north and
            nortwest directions produces better individual, but worse final results. """
        pool = list(range(box_count))
        chromosome = [pool.pop(randrange(box_count))]
        (x, y) = (0, 0)
        for i in range(1, box_count):
            x += 1
            if x == col_count:
                x = 0
                y += 1
                candidates = set()
            else:
                candidates = successors[chromosome[i-1]].intersection(pool)
            chromosome.append(pool.pop(pool.index(choice(tuple(candidates))) if candidates else randrange(len(pool))))
        return Individual(evaluate(chromosome), chromosome)

    def crossover(chromosome_1, chromosome_2):
        """ Produce two children for a given pair of individuals. A random rectangular zone is first selected.
            The corresponding genes in the first individual are copied in the first child. The remaining places
            are filled with the genes of the second individual taken one by one. Symmetrical operations for the
            second child. """
        (x1, y1) = (randrange(col_count), randrange(row_count))
        (x2, y2) = (randrange(x1+1, col_count+1), randrange(y1+1, row_count+1))
        def mate(chromosome_1, chromosome_2):
            used = set(chromosome_1[x+y*col_count] for x in range(x1, x2) for y in range(y1, y2))
            not_used = (allele for allele in chromosome_2 if allele not in used)
            return [chromosome_1[x+y*col_count] if x1 <= x < x2 and y1 <= y < y2 else next(not_used) for y in range(row_count) for x in range(col_count)]
        return (mate(chromosome_1, chromosome_2), mate(chromosome_2, chromosome_1))

    def next_population():
        """ Evolve the population. The best individual is kept. The others are selected by tournament.
            Some selected pairs produce two children. Each selected individual may mutate at certain
            places. Mutation of a gene simply consists in swapping it with another one. """
        result = [best]
        while len(result) < population_size:
            chromosomes = crossover(tournament(), tournament()) if random() < crossover_rate else [tournament()]
            for chromosome in chromosomes:
                for i in range(box_count):
                    if random() < mutation_rate:
                        j = randrange(box_count)
                        (chromosome[i], chromosome[j]) = (chromosome[j], chromosome[i])
                result.append(Individual(evaluate(chromosome), chromosome))
        return result[:population_size]

    def tournament():
        """ Return a COPY of the best chromosome selected among a random sample. """
        return min(sample(population, sample_size)).chromosome[:]
    
    evaluate = fitness(links, multiplicity, col_count, row_count)
    Individual = namedtuple("Individual", ["score", "chromosome"])
    box_count = col_count * row_count
    patience = plateau
    previous_best_score = None
    population = sorted([make_individual() for _ in range(population_size)])
    best = population[0]
    for generation in range(max_generations):
        if best.score == previous_best_score:
            patience -= 1
        else:
            if verbose:
                print("% 3d: %s" % (generation, best.score))
            previous_best_score = best.score
            patience = plateau
        if best.score == (0, 0) or patience == 0 or has_expired():
            # Even if the best individual is not perfect, we stop here
            # without raising an exception.
            break
        population = sorted(next_population())
        best = population[0]
        generation += 1
    result = {
        "distances": best.score[1],
        "crossings": best.score[0],
        "layout": best.chromosome
    }
    mcd.set_layout(**result)
    return mcd.get_clauses()
    
if __name__ == "__main__":
    # python -m mocodo.rewrite._arrange_ga

    from time import time
    import random
    from ..argument_parser import parsed_arguments
    from ..mcd import Mcd

    clauses = """
        SUSPENDISSE: diam
        SOLLICITUDIN, 0N SUSPENDISSE, 0N CONSECTETUER, 0N LOREM: lectus
        CONSECTETUER: elit, sed
        MAECENAS, 1N DIGNISSIM, 1N DIGNISSIM

        DF, 11 LOREM, 1N SUSPENDISSE
        LOREM: ipsum, dolor, sit
        TORTOR, 0N RISUS, 11 DIGNISSIM, 1N CONSECTETUER: nec
        DIGNISSIM: ligula, massa, varius

        DF, 11 RISUS, 0N RISUS
        AMET, 11> LOREM, 01 CONSECTETUER: adipiscing
        RISUS: ultricies, _cras, elementum
        SEMPER, 0N RISUS, 1N DIGNISSIM
    """
    params = parsed_arguments()
    mcd = Mcd(clauses.replace("  ", ""), **params)
    starting_time = time()
    random.seed(42)
    rearrangement = arrange(mcd, {}, lambda: False)
    if rearrangement:
        print()
        mcd.set_layout(**rearrangement)
        print(mcd.get_clauses())
        print()
        print("Cumulated distances:", rearrangement["distances"])
        print("Duration:", time() - starting_time)
        print()
